import os, sys
import torch


class BaseModel():
    def name(self):
        return 'BaseModel'

    def initialize(self, opt):
        self.opt = opt
        self.gpu_ids = opt.gpu_ids
        self.isTrain = opt.isTrain
        self.Tensor = torch.cuda.FloatTensor if self.gpu_ids else torch.Tensor
        self.save_dir = os.path.join(opt.checkpoints_dir, opt.name)

    def set_input(self, input):
        self.input = input

    def forward(self):
        pass

    # used in test time, no backprop
    def test(self):
        pass

    def get_image_paths(self):
        pass

    def optimize_parameters(self):
        pass

    def get_current_visuals(self):
        return self.input

    def get_current_errors(self):
        return {}

    def save(self, label):
        pass

    # helper saving function that can be used by subclasses
    def save_network(self, network, network_label, epoch_label, gpu_ids):
        save_filename = '%s_net_%s.pth' % (epoch_label, network_label)
        save_path = os.path.join(self.save_dir, save_filename)
        torch.save(network.cpu().state_dict(), save_path)
        if len(gpu_ids) and torch.cuda.is_available():
            network.cuda(gpu_ids[0])

    # helper loading function that can be used by subclasses
    def load_network(self, network, network_label, epoch_label,load_path):
        save_filename = '%s_net_%s.pth' % (epoch_label, network_label)
        save_path = os.path.join(self.save_dir, save_filename)
        # #print(save_path)
        network.load_state_dict(torch.load(load_path))

    # update learning rate (called once every epoch)
    def update_learning_rate(self):
        for scheduler in self.schedulers:
            scheduler.step()
        lr = self.optimizers[0].param_groups[0]['lr']
        if (lr <= 0):
            #print('A7A  IS 0 55555  learning rate = %.7f' % lr)
            #print("EXITING TRAINING BECAUSE LR IS <0")
            sys.exit()
        #print('learning rate = %.7f' % lr)



import os.path
import torchvision.transforms as transforms
from data.base_dataset import BaseDataset, get_transform
from data.image_folder import make_dataset
from PIL import Image
import PIL
import random

class UnalignedDataset(BaseDataset):
    def initialize(self, opt):
        self.opt = opt
        self.root = opt.dataroot
        self.dir_A = os.path.join(opt.dataroot, opt.phase + 'A')
        self.dir_B = os.path.join(opt.dataroot, opt.phase + 'B')

        self.A_paths = make_dataset(self.dir_A)
        self.B_paths = make_dataset(self.dir_B)

        self.A_paths = sorted(self.A_paths)
        self.B_paths = sorted(self.B_paths)
        self.A_size = len(self.A_paths)
        self.B_size = len(self.B_paths)
        self.transform = get_transform(opt)

    def __getitem__(self, index):
        A_path = self.A_paths[index % self.A_size]
        index_A = index % self.A_size
        if self.opt.serial_batches:
            index_B = index % self.B_size
        else:
            index_B = random.randint(0, self.B_size - 1)
        B_path = self.B_paths[index_B]
        # #print('(A, B) = (%d, %d)' % (index_A, index_B))
        A_img = Image.open(A_path).convert('RGB')
        B_img = Image.open(B_path).convert('RGB')

        A = self.transform(A_img)
        B = self.transform(B_img)
        if self.opt.which_direction == 'BtoA':
            input_nc = self.opt.output_nc
            output_nc = self.opt.input_nc
        else:
            input_nc = self.opt.input_nc
            output_nc = self.opt.output_nc

        if input_nc == 1:  # RGB to gray
            tmp = A[0, ...] * 0.299 + A[1, ...] * 0.587 + A[2, ...] * 0.114
            A = tmp.unsqueeze(0)

        if output_nc == 1:  # RGB to gray
            tmp = B[0, ...] * 0.299 + B[1, ...] * 0.587 + B[2, ...] * 0.114
            B = tmp.unsqueeze(0)
        return {'A': A, 'B': B,
                'A_paths': A_path, 'B_paths': B_path}

    def __len__(self):
        return max(self.A_size, self.B_size)

    def name(self):
        return 'UnalignedDataset'



import time
from options.train_options import TrainOptions
from data.data_loader import CreateDataLoader
from models.models import create_model
from util.visualizer_time import Visualizer

opt = TrainOptions().parse()
data_loader = CreateDataLoader(opt)
dataset = data_loader.load_data()
dataset_size = len(data_loader)
#print('#training images = %d' % dataset_size)

model = create_model(opt)
visualizer = Visualizer(opt)
total_steps = 0

for epoch in range(opt.epoch_count, opt.niter + opt.niter_decay + 1):
    epoch_start_time = time.time()
    epoch_iter = 0

    for i, data in enumerate(dataset):
        iter_start_time = time.time()
        visualizer.reset()
        total_steps += opt.batchSize
        epoch_iter += opt.batchSize
        model.set_input(data)
        model.optimize_parameters()

        if total_steps % opt.display_freq == 0:
            save_result = total_steps % opt.update_html_freq == 0
            visualizer.display_current_results(model.get_current_visuals(), epoch, save_result)

        if total_steps % opt.#print_freq == 0:
            errors = model.get_current_errors()
            t = (time.time() - iter_start_time) / opt.batchSize
            visualizer.#print_current_errors(epoch, epoch_iter, errors, t)
            if opt.display_id > 0:
                visualizer.plot_current_errors(epoch, float(epoch_iter) / dataset_size, opt, errors)

        if total_steps % opt.save_latest_freq == 0:
            #print('saving the latest model (epoch %d, total_steps %d)' %
                  (epoch, total_steps))
            model.save('latest')

    if epoch % opt.save_epoch_freq == 0:
        #print('saving the model at the end of epoch %d, iters %d' %
              (epoch, total_steps))
        model.save('latest')
        model.save(epoch)

    #print('End of epoch %d / %d \t Time Taken: %d sec' %
          (epoch, opt.niter + opt.niter_decay, time.time() - epoch_start_time))
    model.update_learning_rate()



import os
from .options.test_options import TestOptions
from .data import create_dataset
from .models import create_model
from .util.visualizer import save_images
from .util import html
import requests
from tqdm import tqdm
from ..image_handler import LoadImages
from PIL import Image
from tqdm import tqdm

class CycleGAN:
  
    def __init__(self, model_path=''):
        self.model_name = "cyclegan_model.pth"
        self.is_model_ready = model_path != ''
        self.model_path = model_path
        
        self._check_model(self._get_safe_path(self.model_name))
        
    
    def _get_safe_path(self,filename):
        home_dir = os.path.expanduser("~")
        dir_name = ".lbbnorm"
        safe_path = os.path.join(home_dir, dir_name)
        if not os.path.exists(safe_path):
            os.makedirs(safe_path)
        return os.path.join(safe_path, filename)
    
    def _check_model(self,model_path):
        
        if os.path.exists(model_path):
            self.model_path = model_path
            self.is_model_ready = True
            return True
        else:
            return False
    
    def download_model(self):
        destination = self._get_safe_path(self.model_name)
        
        if self._check_model(destination):
            print('Model already exists.')
            return
        
        download_url = "http://s79.ir/lbb/cyclegan_model.pth"
        response = requests.get(download_url, stream=True)
        total_size_in_bytes= int(response.headers.get('content-length', 0))
        block_size = 1024
        progress_bar = tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)
        with open(destination, 'wb') as file:
            for data in response.iter_content(block_size):
                progress_bar.update(len(data))
                file.write(data)
        progress_bar.close()
        if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
            print("ERROR, something went wrong, check your internet connection.")
            try:
                os.remove(destination)
            except:
                pass
        else:
            self.model_path = destination
            self.is_model_ready = True



    def transform(self, path_image, result_dir='./cyclegan_results/'):
        if not self._check_model(self.model_path) or len(self.model_path) < 1:
            print("""Pretrained model file not found, please download or specify it first.
Hint: you can download the pretrained model with `model.download_model()`, or you can specify it like this `StainGAN(model_path='../../../model.pth')` on creation time.""")
            FileNotFoundError()
            return
        
        images,_,_ = LoadImages(path_image)
        
        image_objects = []
        for img in images:
            image_objects.append(Image.fromarray(img.astype('uint8'), 'RGB'))

        output = []
        # data_root = "assets"
        for img_ in tqdm(image_objects):
            img_ = [img_]
            opt = TestOptions().parse(result_dir)  # get test options
            # hard-code some parameters for test
            opt.num_threads = 0   # test code only supports num_threads = 0
            opt.batch_size = 1    # test code only supports batch_size = 1
            opt.serial_batches = True  # disable data shuffling; comment this line if results on randomly chosen images are needed.
            opt.no_flip = True    # no flip; comment this line if results on flipped images are needed.
            opt.display_id = -1   # no visdom display; the test code saves the results to a HTML file.
            dataset = create_dataset(opt, img_)  # create a dataset given opt.dataset_mode and other options
            model = create_model(opt)      # create a model given opt.model and other options
            model.setup(opt, self.model_path)               # regular setup: load and print networks; create schedulers
            # create a website
            web_dir = os.path.join(opt.results_dir, opt.name, '{}_{}'.format(opt.phase, opt.epoch))  # define the website directory
            if opt.suffix != '':
                web_dir += '_{}'.format(opt.suffix)
            if opt.load_iter > 0:  # load_iter is 0 by default
                web_dir = '{:s}_iter{:d}'.format(web_dir, opt.load_iter)
            #print('creating web directory', web_dir)
            webpage = html.HTML(web_dir, 'Experiment = %s, Phase = %s, Epoch = %s' % (opt.name, opt.phase, opt.epoch))
            # test with eval mode. This only affects layers like batchnorm and dropout.
            # For [pix2pix]: we use batchnorm and dropout in the original pix2pix. You can experiment it with and without eval() mode.
            # For [CycleGAN]: It should not affect CycleGAN as CycleGAN uses instancenorm without dropout.
            if opt.eval:
                model.eval()
            for i, data in enumerate(dataset):
                if i >= opt.num_test:  # only apply our model to opt.num_test images.
                    break
                
                model.set_input(data)  # unpack data from data loader
                model.test()           # run inference
                visuals = model.get_current_visuals()  # get image results
                img_path = model.get_image_paths()     # get image paths
                ims = save_images(webpage, visuals, img_path, aspect_ratio=opt.aspect_ratio, width=opt.display_winsize)
                
            webpage.save()  # save the HTML
            output.append(ims)
            
        if len(output) == 1:
            return output[0]
        else:
            return output


import torch
import itertools
from util.image_pool import ImagePool
from .base_model import BaseModel
from . import networks


class CycleGANModel(BaseModel):
    """
    This class implements the CycleGAN model, for learning image-to-image translation without paired data.

    The model training requires '--dataset_mode unaligned' dataset.
    By default, it uses a '--netG resnet_9blocks' ResNet generator,
    a '--netD basic' discriminator (PatchGAN introduced by pix2pix),
    and a least-square GANs objective ('--gan_mode lsgan').

    CycleGAN paper: https://arxiv.org/pdf/1703.10593.pdf
    """
    @staticmethod
    def modify_commandline_options(parser, is_train=True):
        """Add new dataset-specific options, and rewrite default values for existing options.

        Parameters:
            parser          -- original option parser
            is_train (bool) -- whether training phase or test phase. You can use this flag to add training-specific or test-specific options.

        Returns:
            the modified parser.

        For CycleGAN, in addition to GAN losses, we introduce lambda_A, lambda_B, and lambda_identity for the following losses.
        A (source domain), B (target domain).
        Generators: G_A: A -> B; G_B: B -> A.
        Discriminators: D_A: G_A(A) vs. B; D_B: G_B(B) vs. A.
        Forward cycle loss:  lambda_A * ||G_B(G_A(A)) - A|| (Eqn. (2) in the paper)
        Backward cycle loss: lambda_B * ||G_A(G_B(B)) - B|| (Eqn. (2) in the paper)
        Identity loss (optional): lambda_identity * (||G_A(B) - B|| * lambda_B + ||G_B(A) - A|| * lambda_A) (Sec 5.2 "Photo generation from paintings" in the paper)
        Dropout is not used in the original CycleGAN paper.
        """
        parser.set_defaults(no_dropout=True)  # default CycleGAN did not use dropout
        if is_train:
            parser.add_argument('--lambda_A', type=float, default=10.0, help='weight for cycle loss (A -> B -> A)')
            parser.add_argument('--lambda_B', type=float, default=10.0, help='weight for cycle loss (B -> A -> B)')
            parser.add_argument('--lambda_identity', type=float, default=0.5, help='use identity mapping. Setting lambda_identity other than 0 has an effect of scaling the weight of the identity mapping loss. For example, if the weight of the identity loss should be 10 times smaller than the weight of the reconstruction loss, please set lambda_identity = 0.1')

        return parser

    def __init__(self, opt):
        """Initialize the CycleGAN class.

        Parameters:
            opt (Option class)-- stores all the experiment flags; needs to be a subclass of BaseOptions
        """
        BaseModel.__init__(self, opt)
        # specify the training losses you want to print out. The training/test scripts will call <BaseModel.get_current_losses>
        self.loss_names = ['D_A', 'G_A', 'cycle_A', 'idt_A', 'D_B', 'G_B', 'cycle_B', 'idt_B']
        # specify the images you want to save/display. The training/test scripts will call <BaseModel.get_current_visuals>
        visual_names_A = ['real_A', 'fake_B', 'rec_A']
        visual_names_B = ['real_B', 'fake_A', 'rec_B']
        if self.isTrain and self.opt.lambda_identity > 0.0:  # if identity loss is used, we also visualize idt_B=G_A(B) ad idt_A=G_A(B)
            visual_names_A.append('idt_B')
            visual_names_B.append('idt_A')

        self.visual_names = visual_names_A + visual_names_B  # combine visualizations for A and B
        # specify the models you want to save to the disk. The training/test scripts will call <BaseModel.save_networks> and <BaseModel.load_networks>.
        if self.isTrain:
            self.model_names = ['G_A', 'G_B', 'D_A', 'D_B']
        else:  # during test time, only load Gs
            self.model_names = ['G_A', 'G_B']

        # define networks (both Generators and discriminators)
        # The naming is different from those used in the paper.
        # Code (vs. paper): G_A (G), G_B (F), D_A (D_Y), D_B (D_X)
        self.netG_A = networks.define_G(opt.input_nc, opt.output_nc, opt.ngf, opt.netG, opt.norm,
                                        not opt.no_dropout, opt.init_type, opt.init_gain, self.gpu_ids)
        self.netG_B = networks.define_G(opt.output_nc, opt.input_nc, opt.ngf, opt.netG, opt.norm,
                                        not opt.no_dropout, opt.init_type, opt.init_gain, self.gpu_ids)

        if self.isTrain:  # define discriminators
            self.netD_A = networks.define_D(opt.output_nc, opt.ndf, opt.netD,
                                            opt.n_layers_D, opt.norm, opt.init_type, opt.init_gain, self.gpu_ids)
            self.netD_B = networks.define_D(opt.input_nc, opt.ndf, opt.netD,
                                            opt.n_layers_D, opt.norm, opt.init_type, opt.init_gain, self.gpu_ids)

        if self.isTrain:
            if opt.lambda_identity > 0.0:  # only works when input and output images have the same number of channels
                assert(opt.input_nc == opt.output_nc)
            self.fake_A_pool = ImagePool(opt.pool_size)  # create image buffer to store previously generated images
            self.fake_B_pool = ImagePool(opt.pool_size)  # create image buffer to store previously generated images
            # define loss functions
            self.criterionGAN = networks.GANLoss(opt.gan_mode).to(self.device)  # define GAN loss.
            self.criterionCycle = torch.nn.L1Loss()
            self.criterionIdt = torch.nn.L1Loss()
            # initialize optimizers; schedulers will be automatically created by function <BaseModel.setup>.
            self.optimizer_G = torch.optim.Adam(itertools.chain(self.netG_A.parameters(), self.netG_B.parameters()), lr=opt.lr, betas=(opt.beta1, 0.999))
            self.optimizer_D = torch.optim.Adam(itertools.chain(self.netD_A.parameters(), self.netD_B.parameters()), lr=opt.lr, betas=(opt.beta1, 0.999))
            self.optimizers.append(self.optimizer_G)
            self.optimizers.append(self.optimizer_D)

    def set_input(self, input):
        """Unpack input data from the dataloader and perform necessary pre-processing steps.

        Parameters:
            input (dict): include the data itself and its metadata information.

        The option 'direction' can be used to swap domain A and domain B.
        """
        AtoB = self.opt.direction == 'AtoB'
        # self.real_A = input['A' if AtoB else 'B'].to(self.device)
        # self.real_B = input['B' if AtoB else 'A'].to(self.device)
        self.real_A = input['A' if AtoB else 'B']
        self.real_B = input['B' if AtoB else 'A']
        
        self.image_paths = input['A_paths' if AtoB else 'B_paths']

    def forward(self):
        """Run forward pass; called by both functions <optimize_parameters> and <test>."""
        self.fake_B = self.netG_A(self.real_A)  # G_A(A)
        self.rec_A = self.netG_B(self.fake_B)   # G_B(G_A(A))
        self.fake_A = self.netG_B(self.real_B)  # G_B(B)
        self.rec_B = self.netG_A(self.fake_A)   # G_A(G_B(B))

    def backward_D_basic(self, netD, real, fake):
        """Calculate GAN loss for the discriminator

        Parameters:
            netD (network)      -- the discriminator D
            real (tensor array) -- real images
            fake (tensor array) -- images generated by a generator

        Return the discriminator loss.
        We also call loss_D.backward() to calculate the gradients.
        """
        # Real
        pred_real = netD(real)
        loss_D_real = self.criterionGAN(pred_real, True)
        # Fake
        pred_fake = netD(fake.detach())
        loss_D_fake = self.criterionGAN(pred_fake, False)
        # Combined loss and calculate gradients
        loss_D = (loss_D_real + loss_D_fake) * 0.5
        loss_D.backward()
        return loss_D

    def backward_D_A(self):
        """Calculate GAN loss for discriminator D_A"""
        fake_B = self.fake_B_pool.query(self.fake_B)
        self.loss_D_A = self.backward_D_basic(self.netD_A, self.real_B, fake_B)

    def backward_D_B(self):
        """Calculate GAN loss for discriminator D_B"""
        fake_A = self.fake_A_pool.query(self.fake_A)
        self.loss_D_B = self.backward_D_basic(self.netD_B, self.real_A, fake_A)

    def backward_G(self):
        """Calculate the loss for generators G_A and G_B"""
        lambda_idt = self.opt.lambda_identity
        lambda_A = self.opt.lambda_A
        lambda_B = self.opt.lambda_B
        # Identity loss
        if lambda_idt > 0:
            # G_A should be identity if real_B is fed: ||G_A(B) - B||
            self.idt_A = self.netG_A(self.real_B)
            self.loss_idt_A = self.criterionIdt(self.idt_A, self.real_B) * lambda_B * lambda_idt
            # G_B should be identity if real_A is fed: ||G_B(A) - A||
            self.idt_B = self.netG_B(self.real_A)
            self.loss_idt_B = self.criterionIdt(self.idt_B, self.real_A) * lambda_A * lambda_idt
        else:
            self.loss_idt_A = 0
            self.loss_idt_B = 0

        # GAN loss D_A(G_A(A))
        self.loss_G_A = self.criterionGAN(self.netD_A(self.fake_B), True)
        # GAN loss D_B(G_B(B))
        self.loss_G_B = self.criterionGAN(self.netD_B(self.fake_A), True)
        # Forward cycle loss || G_B(G_A(A)) - A||
        self.loss_cycle_A = self.criterionCycle(self.rec_A, self.real_A) * lambda_A
        # Backward cycle loss || G_A(G_B(B)) - B||
        self.loss_cycle_B = self.criterionCycle(self.rec_B, self.real_B) * lambda_B
        # combined loss and calculate gradients
        self.loss_G = self.loss_G_A + self.loss_G_B + self.loss_cycle_A + self.loss_cycle_B + self.loss_idt_A + self.loss_idt_B
        self.loss_G.backward()

    def optimize_parameters(self):
        """Calculate losses, gradients, and update network weights; called in every training iteration"""
        # forward
        self.forward()      # compute fake images and reconstruction images.
        # G_A and G_B
        self.set_requires_grad([self.netD_A, self.netD_B], False)  # Ds require no gradients when optimizing Gs
        self.optimizer_G.zero_grad()  # set G_A and G_B's gradients to zero
        self.backward_G()             # calculate gradients for G_A and G_B
        self.optimizer_G.step()       # update G_A and G_B's weights
        # D_A and D_B
        self.set_requires_grad([self.netD_A, self.netD_B], True)
        self.optimizer_D.zero_grad()   # set D_A and D_B's gradients to zero
        self.backward_D_A()      # calculate gradients for D_A
        self.backward_D_B()      # calculate graidents for D_B
        self.optimizer_D.step()  # update D_A and D_B's weights



import os
from data.base_dataset import BaseDataset, get_params, get_transform
from data.image_folder import make_dataset
from PIL import Image


class AlignedDataset(BaseDataset):
    """A dataset class for paired image dataset.

    It assumes that the directory '/path/to/data/train' contains image pairs in the form of {A,B}.
    During test time, you need to prepare a directory '/path/to/data/test'.
    """

    def __init__(self, opt):
        """Initialize this dataset class.

        Parameters:
            opt (Option class) -- stores all the experiment flags; needs to be a subclass of BaseOptions
        """
        BaseDataset.__init__(self, opt)
        self.dir_AB = os.path.join(opt.dataroot, opt.phase)  # get the image directory
        self.AB_paths = sorted(make_dataset(self.dir_AB, opt.max_dataset_size))  # get image paths
        assert(self.opt.load_size >= self.opt.crop_size)   # crop_size should be smaller than the size of loaded image
        self.input_nc = self.opt.output_nc if self.opt.direction == 'BtoA' else self.opt.input_nc
        self.output_nc = self.opt.input_nc if self.opt.direction == 'BtoA' else self.opt.output_nc

    def __getitem__(self, index):
        """Return a data point and its metadata information.

        Parameters:
            index - - a random integer for data indexing

        Returns a dictionary that contains A, B, A_paths and B_paths
            A (tensor) - - an image in the input domain
            B (tensor) - - its corresponding image in the target domain
            A_paths (str) - - image paths
            B_paths (str) - - image paths (same as A_paths)
        """
        # read a image given a random integer index
        AB_path = self.AB_paths[index]
        AB = Image.open(AB_path).convert('RGB')
        # split AB image into A and B
        w, h = AB.size
        w2 = int(w / 2)
        A = AB.crop((0, 0, w2, h))
        B = AB.crop((w2, 0, w, h))

        # apply the same transform to both A and B
        transform_params = get_params(self.opt, A.size)
        A_transform = get_transform(self.opt, transform_params, grayscale=(self.input_nc == 1))
        B_transform = get_transform(self.opt, transform_params, grayscale=(self.output_nc == 1))

        A = A_transform(A)
        B = B_transform(B)

        return {'A': A, 'B': B, 'A_paths': AB_path, 'B_paths': AB_path}

    def __len__(self):
        """Return the total number of images in the dataset."""
        return len(self.AB_paths)



import torch.utils.data as data
from PIL import Image
import torchvision.transforms as transforms

class BaseDataset(data.Dataset):
    def __init__(self):
        super(BaseDataset, self).__init__()

    def name(self):
        return 'BaseDataset'

    def initialize(self, opt):
        pass

def get_transform(opt):
    transform_list = []
    if opt.resize_or_crop == 'resize_and_crop':
        osize = [opt.loadSize, opt.loadSize]
        transform_list.append(transforms.Resize(osize, Image.BICUBIC))
        transform_list.append(transforms.RandomCrop(opt.fineSize))
    elif opt.resize_or_crop == 'crop':
        transform_list.append(transforms.RandomCrop(opt.fineSize))
    elif opt.resize_or_crop == 'scale_width':
        transform_list.append(transforms.Lambda(
            lambda img: __scale_width(img, opt.fineSize)))
    elif opt.resize_or_crop == 'scale_width_and_crop':
        transform_list.append(transforms.Lambda(
            lambda img: __scale_width(img, opt.loadSize)))
        transform_list.append(transforms.RandomCrop(opt.fineSize))

    if opt.isTrain and not opt.no_flip:
        transform_list.append(transforms.RandomHorizontalFlip())

    transform_list += [transforms.ToTensor(),
                       transforms.Normalize((0.5, 0.5, 0.5),
                                            (0.5, 0.5, 0.5))]
    return transforms.Compose(transform_list)

def __scale_width(img, target_width):
    ow, oh = img.size
    if (ow == target_width):
        return img
    w = target_width
    h = int(target_width * oh / ow)
    return img.resize((w, h), Image.BICUBIC)



from .preprocess import RandStainPreprocess
from .convertor import RandStainNAConvertor
import cv2
from ..image_handler import LoadImages



class RandStain():
    def __init__(self, yaml_file=None) -> None:
        self.yaml_file = yaml_file
    
    def fit(self, data_dir='', save_dir='./', dataset_name="", methods="", color_space="LAB", random=False, n=0):
        
        self.yaml_file = RandStainPreprocess().do( data_dir, save_dir, dataset_name, methods, color_space, random, n)
        
    def transform(self,image_paths,yaml_file=None,
            std_hyper = 0.0,
            distribution = 'normal',
            probability = 1.0,
            is_train = False
                  ):
        
        if yaml_file == None and self.yaml_file == None:
            print('You have to fit the model first.')
            AssertionError()
            return
        if yaml_file != None:
            self.yaml_file = yaml_file
        
        _,_,image_paths_list = LoadImages(image_paths)
        
        
        randstainna = RandStainNAConvertor(
            yaml_file = self.yaml_file,
            std_hyper = std_hyper,
            distribution = distribution,
            probability = probability,
            is_train = is_train
        )
        
        output = []
        for img_path in image_paths_list:
            img = randstainna(cv2.imread(img_path))
            output.append(img)
        
        if len(output) == 1:
            return output[0]
        else:
            return output
        
        
        
        
        


from cyclegan import CycleGAN
from PIL import Image
import matplotlib.pyplot as plt

load_path = "D:\\research\\articles\\code_template\\normalization\\cyclegan - static-checkpoint\\CycleGAN\\results\\mitos-atypia\\latest_net_G_A.pth"
path_image = "D:\\research\\articles\\code_template\\normalization\\cyclegan - static-checkpoint\\CycleGAN\\assets\\testA\\1001702_1.tif"
result_dir = "result-final-images"
image =Image.open(path_image)
result = CycleGAN().transform(load_path, image, result_dir)
normalized_image = Image.fromarray(result)

plt.imshow(normalized_image)
plt.show()



import numpy as np
import torch
import os
import matplotlib.pyplot as plt
from collections import OrderedDict
from torch.autograd import Variable
import itertools
import util.util as util
from util.image_pool import ImagePool
from .base_model import BaseModel
from . import networks
import sys


class CycleGANModel(BaseModel):
    def name(self):
        return 'CycleGANModel'

    def initialize(self, opt):
        BaseModel.initialize(self, opt)

        nb = opt.batchSize
        size = opt.fineSize

        # load/define networks
        # The naming conversion is different from those used in the paper
        # Code (paper): G_A (G), G_B (F), D_A (D_Y), D_B (D_X)

        self.netG_A = networks.define_G(opt.input_nc, opt.output_nc,
                                        opt.ngf, opt.which_model_netG, opt.norm, not opt.no_dropout, opt.init_type,
                                        self.gpu_ids)
        self.netG_B = networks.define_G(opt.output_nc, opt.input_nc,
                                        opt.ngf, opt.which_model_netG, opt.norm, not opt.no_dropout, opt.init_type,
                                        self.gpu_ids)

        if self.isTrain:
            use_sigmoid = opt.no_lsgan
            self.netD_A = networks.define_D(opt.output_nc, opt.ndf,
                                            opt.which_model_netD,
                                            opt.n_layers_D, opt.norm, use_sigmoid, opt.init_type, self.gpu_ids)
            self.netD_B = networks.define_D(opt.input_nc, opt.ndf,
                                            opt.which_model_netD,
                                            opt.n_layers_D, opt.norm, use_sigmoid, opt.init_type, self.gpu_ids)
        if not self.isTrain or opt.continue_train:
            which_epoch = opt.which_epoch
            self.load_network(self.netG_A, 'G_A', which_epoch)
            self.load_network(self.netG_B, 'G_B', which_epoch)
            if self.isTrain:
                self.load_network(self.netD_A, 'D_A', which_epoch)
                self.load_network(self.netD_B, 'D_B', which_epoch)

        if self.isTrain:
            self.old_lr = opt.lr
            self.fake_A_pool = ImagePool(opt.pool_size)
            self.fake_B_pool = ImagePool(opt.pool_size)
            # define loss functions
            self.criterionGAN = networks.GANLoss(use_lsgan=not opt.no_lsgan, tensor=self.Tensor)
            self.criterionCycle = torch.nn.L1Loss()
            self.criterionIdt = torch.nn.L1Loss()
            # gradient loss as per aicha
            if opt.aicha_loss:
                self.criterionAicha = torch.nn.L1Loss()

            # initialize optimizers
            self.optimizer_G = torch.optim.Adam(itertools.chain(self.netG_A.parameters(), self.netG_B.parameters()),
                                                lr=opt.lr, betas=(opt.beta1, 0.999))
            self.optimizer_D_A = torch.optim.Adam(self.netD_A.parameters(), lr=opt.lr, betas=(opt.beta1, 0.999))
            self.optimizer_D_B = torch.optim.Adam(self.netD_B.parameters(), lr=opt.lr, betas=(opt.beta1, 0.999))

            self.optimizers = []
            self.schedulers = []
            self.optimizers.append(self.optimizer_G)
            self.optimizers.append(self.optimizer_D_A)
            self.optimizers.append(self.optimizer_D_B)
            for optimizer in self.optimizers:
                self.schedulers.append(networks.get_scheduler(optimizer, opt))

        # #print('---------- Networks initialized -------------')
        # networks.#print_network(self.netG_A)
        # networks.#print_network(self.netG_B)
        # if self.isTrain:
        #     networks.#print_network(self.netD_A)
        #     networks.#print_network(self.netD_B)
        # #print('-----------------------------------------------')

    def set_input(self, input):
        AtoB = self.opt.which_direction == 'AtoB'
        input_A = input['A' if AtoB else 'B']
        input_B = input['B' if AtoB else 'A']
        # if len(self.gpu_ids) > 0:
        #     input_A = input_A.cuda(self.gpu_ids[0], async=True)
        #     input_B = input_B.cuda(self.gpu_ids[0], async=True)
        self.input_A = input_A
        self.input_B = input_B
        self.image_paths = input['A_paths' if AtoB else 'B_paths']

    def forward(self):
        self.real_A = Variable(self.input_A)
        self.real_B = Variable(self.input_B)

    def test(self):
        real_A = Variable(self.input_A, volatile=True)
        fake_B = self.netG_A(real_A)
        self.rec_A = self.netG_B(fake_B).data
        self.fake_B = fake_B.data

        real_B = Variable(self.input_B, volatile=True)
        fake_A = self.netG_B(real_B)
        self.rec_B = self.netG_A(fake_A).data
        self.fake_A = fake_A.data

    # get image paths
    def get_image_paths(self):
        return self.image_paths

    def backward_D_basic(self, netD, real, fake):
        # Real
        pred_real = netD(real)
        loss_D_real = self.criterionGAN(pred_real, True)
        # Fake
        pred_fake = netD(fake.detach())
        loss_D_fake = self.criterionGAN(pred_fake, False)
        # Combined loss
        loss_D = (loss_D_real + loss_D_fake) * 0.5
        # backward
        loss_D.backward()
        return loss_D

    def backward_D_A(self):
        fake_B = self.fake_B_pool.query(self.fake_B)
        loss_D_A = self.backward_D_basic(self.netD_A, self.real_B, fake_B)
        self.loss_D_A = loss_D_A.data[0]

    def backward_D_B(self):
        fake_A = self.fake_A_pool.query(self.fake_A)
        loss_D_B = self.backward_D_basic(self.netD_B, self.real_A, fake_A)
        self.loss_D_B = loss_D_B.data[0]

    def backward_G(self):
        lambda_idt = self.opt.identity
        lambda_A = self.opt.lambda_A
        lambda_B = self.opt.lambda_B
        # Identity loss
        if lambda_idt > 0:
            # G_A should be identity if real_B is fed.
            idt_A = self.netG_A(self.real_B)
            loss_idt_A = self.criterionIdt(idt_A, self.real_B) * lambda_B * lambda_idt
            # G_B should be identity if real_A is fed.
            idt_B = self.netG_B(self.real_A)
            loss_idt_B = self.criterionIdt(idt_B, self.real_A) * lambda_A * lambda_idt

            self.idt_A = idt_A.data
            self.idt_B = idt_B.data
            self.loss_idt_A = loss_idt_A.data[0]
            self.loss_idt_B = loss_idt_B.data[0]
        else:
            loss_idt_A = 0
            loss_idt_B = 0
            self.loss_idt_A = 0
            self.loss_idt_B = 0

        # GAN loss D_A(G_A(A))
        fake_B = self.netG_A(self.real_A)
        pred_fake = self.netD_A(fake_B)
        loss_G_A = self.criterionGAN(pred_fake, True)

        # GAN loss D_B(G_B(B))
        fake_A = self.netG_B(self.real_B)
        pred_fake = self.netD_B(fake_A)
        loss_G_B = self.criterionGAN(pred_fake, True)

        # Forward cycle loss
        rec_A = self.netG_B(fake_B)
        loss_cycle_A = self.criterionCycle(rec_A, self.real_A) * lambda_A
        # #print("loss_cycle_A  ",loss_cycle_A.grad)

        # Backward cycle loss
        rec_B = self.netG_A(fake_A)
        loss_cycle_B = self.criterionCycle(rec_B, self.real_B) * lambda_B

        # aicha forward
        if self.opt.aicha_loss:
            #print("I AM AICHCA LOSS")
            # w1 = Variable(torch.Tensor([1.0, 2.0, 3.0]), requires_grad=True)
            #
            # #print(w1.gra)
            #
            # sys.exit()
            #
            # realA_img = util.tensor2im(self.real_A.data)
            # recA_img = util.tensor2im(rec_A.data)
            # #print(realA_img.shape)
            # #print("---------")
            # gradient_input = self.get_gradient(recA_img)
            #
            # # #print(gradient_input)
            #
            # wx = Variable(torch.from_numpy(np.multiply(gradient_input, realA_img)), requires_grad=True)
            # wg = Variable(torch.from_numpy(np.multiply(gradient_input, recA_img)), requires_grad=True)
            # loss_Aicha_A = self.criterionAicha(wx, wg) * self.opt.lambda_Aicha

            # aicha backwards

            realB_img = util.tensor2im(self.real_B.data)
            recB_img = util.tensor2im(rec_B.data)
            plt.imshow(realB_img);
            plt.imshow(recB_img);
            plt.show()

            sys.exit()

            gradient_input = np.gradient(realB_img)
            # #print("gradient_input", gradient_input.size)
            gradient_input = torch.from_numpy(gradient_input)
            sys.exit()
            grad_times_real = torch.mul(gradient_input, self.real_B)

            # wx = Variable(torch.from_numpy(), requires_grad=True)
            wx = grad_times_real
            #print('grad_times_real', grad_times_real)
            sys.exit()
            wg = Variable(torch.from_numpy(np.multiply(gradient_input, recB_img)), requires_grad=True)
            loss_Aicha_B = self.criterionAicha(wx, wg) * self.opt.lambda_Aicha

            # aicha combined loss

            loss_G = loss_G_A + loss_G_B + loss_Aicha_A + loss_Aicha_B + loss_idt_A + loss_idt_B
        else:
            # combined loss
            loss_G = loss_G_A + loss_G_B + loss_cycle_A + loss_cycle_B + loss_idt_A + loss_idt_B

        loss_G.backward()

        self.fake_B = fake_B.data
        self.fake_A = fake_A.data
        self.rec_A = rec_A.data
        self.rec_B = rec_B.data

        self.loss_G_A = loss_G_A.data[0]
        self.loss_G_B = loss_G_B.data[0]

        if self.opt.aicha_loss:
            # AichaLoss
            self.loss_Aicha_A = loss_Aicha_A.data[0]
            self.loss_Aicha_B = loss_Aicha_B.data[0]
        else:
            # cycleLoss
            self.loss_cycle_A = loss_cycle_A.data[0]
            self.loss_cycle_B = loss_cycle_B.data[0]

    #
    # def get_gradient(self, img):
    #     # #print("get_gradient ",img)
    #     # return np.gradient(np.array(img))
    #     return np.gradient(img)

    def optimize_parameters(self):
        # forward
        self.forward()
        # G_A and G_B
        self.optimizer_G.zero_grad()
        self.backward_G()
        self.optimizer_G.step()
        # D_A
        self.optimizer_D_A.zero_grad()
        self.backward_D_A()
        self.optimizer_D_A.step()
        # D_B
        self.optimizer_D_B.zero_grad()
        self.backward_D_B()
        self.optimizer_D_B.step()

    def get_current_errors(self):
        ret_errors = OrderedDict([('D_A', self.loss_D_A), ('G_A', self.loss_G_A), ('Cyc_A', self.loss_cycle_A),
                                  ('D_B', self.loss_D_B), ('G_B', self.loss_G_B), ('Cyc_B', self.loss_cycle_B)])
        # aicha
        # ret_errors = OrderedDict([('D_A', self.loss_D_A), ('G_A', self.loss_G_A), ('Aicha_A', self.loss_cycle_A),
        #                           ('D_B', self.loss_D_B), ('G_B', self.loss_G_B), ('Aicha_B', self.loss_cycle_B)])

        if self.opt.identity > 0.0:
            ret_errors['idt_A'] = self.loss_idt_A
            ret_errors['idt_B'] = self.loss_idt_B
        return ret_errors

    def get_current_visuals(self):
        real_A = util.tensor2im(self.input_A)
        fake_B = util.tensor2im(self.fake_B)
        rec_A = util.tensor2im(self.rec_A)
        real_B = util.tensor2im(self.input_B)
        fake_A = util.tensor2im(self.fake_A)
        rec_B = util.tensor2im(self.rec_B)
        ret_visuals = OrderedDict([('real_A', real_A), ('fake_B', fake_B), ('rec_A', rec_A),
                                   ('real_B', real_B), ('fake_A', fake_A), ('rec_B', rec_B)])
        if self.opt.isTrain and self.opt.identity > 0.0:
            ret_visuals['idt_A'] = util.tensor2im(self.idt_A)
            ret_visuals['idt_B'] = util.tensor2im(self.idt_B)
        return ret_visuals

    def save(self, label):
        self.save_network(self.netG_A, 'G_A', label, self.gpu_ids)
        self.save_network(self.netD_A, 'D_A', label, self.gpu_ids)
        self.save_network(self.netG_B, 'G_B', label, self.gpu_ids)
        self.save_network(self.netD_B, 'D_B', label, self.gpu_ids)



"""General-purpose test script for image-to-image translation.

Once you have trained your model with train.py, you can use this script to test the model.
It will load a saved model from '--checkpoints_dir' and save the results to '--results_dir'.

It first creates model and dataset given the option. It will hard-code some parameters.
It then runs inference for '--num_test' images and save results to an HTML file.

Example (You need to train models first or download pre-trained models from our website):
    Test a CycleGAN model (both sides):
        python test.py --dataroot ./datasets/maps --name maps_cyclegan --model cycle_gan

    Test a CycleGAN model (one side only):
        python test.py --dataroot datasets/horse2zebra/testA --name horse2zebra_pretrained --model test --no_dropout

    The option '--model test' is used for generating CycleGAN results only for one side.
    This option will automatically set '--dataset_mode single', which only loads the images from one set.
    On the contrary, using '--model cycle_gan' requires loading and generating results in both directions,
    which is sometimes unnecessary. The results will be saved at ./results/.
    Use '--results_dir <directory_path_to_save_result>' to specify the results directory.

    Test a pix2pix model:
        python test.py --dataroot ./datasets/facades --name facades_pix2pix --model pix2pix --direction BtoA

See options/base_options.py and options/test_options.py for more test options.
See training and test tips at: https://github.com/junyanz/pytorch-CycleGAN-and-pix2pix/blob/master/docs/tips.md
See frequently asked questions at: https://github.com/junyanz/pytorch-CycleGAN-and-pix2pix/blob/master/docs/qa.md
"""
import os
from options.test_options import TestOptions
from data import create_dataset
from models import create_model
from util.visualizer import save_images
from util import html


if __name__ == '__main__':
    # data_root = "assets"
    load_path = ""
    result_dir = ""
    opt = TestOptions().parse(result_dir)  # get test options
    # hard-code some parameters for test
    opt.num_threads = 0   # test code only supports num_threads = 0
    opt.batch_size = 1    # test code only supports batch_size = 1
    opt.serial_batches = True  # disable data shuffling; comment this line if results on randomly chosen images are needed.
    opt.no_flip = True    # no flip; comment this line if results on flipped images are needed.
    opt.display_id = -1   # no visdom display; the test code saves the results to a HTML file.
    dataset = create_dataset(opt)  # create a dataset given opt.dataset_mode and other options
    model = create_model(opt)      # create a model given opt.model and other options
    model.setup(opt, load_path)               # regular setup: load and print networks; create schedulers
    # create a website
    web_dir = os.path.join(opt.results_dir, opt.name, '{}_{}'.format(opt.phase, opt.epoch))  # define the website directory
    if opt.suffix != '':
        web_dir += '_{}'.format(opt.suffix)
    if opt.load_iter > 0:  # load_iter is 0 by default
        web_dir = '{:s}_iter{:d}'.format(web_dir, opt.load_iter)
    #print('creating web directory', web_dir)
    webpage = html.HTML(web_dir, 'Experiment = %s, Phase = %s, Epoch = %s' % (opt.name, opt.phase, opt.epoch))
    # test with eval mode. This only affects layers like batchnorm and dropout.
    # For [pix2pix]: we use batchnorm and dropout in the original pix2pix. You can experiment it with and without eval() mode.
    # For [CycleGAN]: It should not affect CycleGAN as CycleGAN uses instancenorm without dropout.
    if opt.eval:
        model.eval()
    for i, data in enumerate(dataset):
        if i >= opt.num_test:  # only apply our model to opt.num_test images.
            break
        
        model.set_input(data)  # unpack data from data loader
        model.test()           # run inference
        visuals = model.get_current_visuals()  # get image results
        img_path = model.get_image_paths()     # get image paths
        if i % 5 == 0:  # save images to an HTML file
            ...
            #print('processing (%04d)-th image... %s' % (i, img_path))
        save_images(webpage, visuals, img_path, aspect_ratio=opt.aspect_ratio, width=opt.display_winsize)
    webpage.save()  # save the HTML


import importlib
from .macenko import Macenko
from .vahadane import Vahadane
from .reinhard import Reinhard
from .modified_reinhard import ModifiedReinhard
from .adaptive_color_deconvolution import AdaptiveColorDeconvolution
from .staingan.staingan import StainGAN
from .cyclegan.cyclegan import CycleGAN
from .stainnet.stain_net import StainNet
from .randstain.randstainna import RandStain


#from .slide_patcher import SlidePatcher


# def __getattr__(name):
#     if name == 'Macenko':
#         sub_class_module = importlib.import_module('.macenko', __name__)
#         sub_class = getattr(sub_class_module, name)
#         return sub_class
#     elif name == 'Vahadane':
#         sub_class_module = importlib.import_module('.vahadane', __name__)
#         sub_class = getattr(sub_class_module, name)
#         return sub_class
#     elif name == 'AdaptiveColorDeconvolution':
#         sub_class_module = importlib.import_module('.adaptive_color_deconvolution', __name__)
#         sub_class = getattr(sub_class_module, name)
#         return sub_class
#     elif name == 'Reinhard':
#         sub_class_module = importlib.import_module('.reinhard', __name__)
#         sub_class = getattr(sub_class_module, name)
#         return sub_class
#     elif name == 'ModifiedReinhard':
#         sub_class_module = importlib.import_module('.modified_reinhard', __name__)
#         sub_class = getattr(sub_class_module, name)
#         return sub_class
#     elif name == 'StainGAN':
#         sub_class_module = importlib.import_module('.staingan.staingan', __name__)
#         sub_class = getattr(sub_class_module, name)
#         return sub_class
#     elif name == 'RandstainNA':
#         sub_class_module = importlib.import_module('.randstain', __name__)
#         sub_class = getattr(sub_class_module, name)
#         return sub_class
    
#     raise AttributeError(f"module {__name__} has no attribute {name}")


import argparse
import os
from ..util import util
import torch

class BaseOptions():
    def __init__(self):
        self.parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        self.initialized = False

    def initialize(self, path_size):

        self.parser.add_argument('--dataroot', default='', help='path to images (should have subfolders trainA, trainB, valA, valB, etc)')
        self.parser.add_argument('--batchSize', type=int, default=4, help='input batch size')
        self.parser.add_argument('--loadSize', type=int, default=path_size, help='scale images to this size')
        self.parser.add_argument('--fineSize', type=int, default=path_size, help='then crop to this size')
        self.parser.add_argument('--input_nc', type=int, default=3, help='# of input image channels')
        self.parser.add_argument('--output_nc', type=int, default=3, help='# of output image channels')
        self.parser.add_argument('--ngf', type=int, default=64, help='# of gen filters in first conv layer')
        self.parser.add_argument('--ndf', type=int, default=64, help='# of discrim filters in first conv layer')
        self.parser.add_argument('--which_model_netD', type=str, default='basic', help='selects model to use for netD')
        self.parser.add_argument('--which_model_netG', type=str, default='resnet_9blocks', help='selects model to use for netG')
        self.parser.add_argument('--n_layers_D', type=int, default=3, help='only used if which_model_netD==n_layers')
        self.parser.add_argument('--gpu_ids', type=str, default='-1', help='gpu ids: e.g. 0  0,1,2, 0,2. use -1 for CPU')
        self.parser.add_argument('--name', type=str, default='experiment_name', help='name of the experiment. It decides where to store samples and models')
        self.parser.add_argument('--dataset_mode', type=str, default='single', help='chooses how datasets are loaded. [unaligned | aligned | single]')
        # self.parser.add_argument('--model', type=str, default='test',help='chooses which model to use. cycle_gan, pix2pix, test')
        self.parser.add_argument('--which_direction', type=str, default='AtoB', help='AtoB or BtoA')
        self.parser.add_argument('--nThreads', default=2, type=int, help='# threads for loading data')
        self.parser.add_argument('--checkpoints_dir', type=str, default='./checkpoints/aicha/', help='models are saved here')
        self.parser.add_argument('--norm', type=str, default='instance', help='instance normalization or batch normalization')
        self.parser.add_argument('--serial_batches', action='store_true', help='if true, takes images in order to make batches, otherwise takes them randomly')
        self.parser.add_argument('--display_winsize', type=int, default=256,  help='display window size')
        self.parser.add_argument('--display_id', type=int, default=1, help='window id of the web display')
        self.parser.add_argument('--display_port', type=int, default=8097, help='visdom port of the web display')
        self.parser.add_argument('--no_dropout', action='store_false', help='no dropout for the generator')
        self.parser.add_argument('--max_dataset_size', type=int, default=float("inf"), help='Maximum number of samples allowed per dataset. If the dataset directory contains more than max_dataset_size, only a subset is loaded.')
        self.parser.add_argument('--resize_or_crop', type=str, default='resize_and_crop', help='scaling and cropping of images at load time [resize_and_crop|crop|scale_width|scale_width_and_crop]')
        self.parser.add_argument('--no_flip', action='store_true', help='if specified, do not flip the images for data augmentation')
        self.parser.add_argument('--init_type', type=str, default='normal', help='network initialization [normal|xavier|kaiming|orthogonal]')

        self.initialized = True

    def parse(self, path_size,results_dir):
        if not self.initialized:
            self.initialize(path_size,results_dir)
        # self.opt = self.parser.parse_args()
        self.opt, unknown = self.parser.parse_known_args()
        self.opt.isTrain = self.isTrain   # train or test

        str_ids = self.opt.gpu_ids.split(',')
        self.opt.gpu_ids = []
        for str_id in str_ids:
            id = int(str_id)
            if id >= 0:
                self.opt.gpu_ids.append(id)

        # set gpu ids
        # if len(self.opt.gpu_ids) > 0:
        #     torch.cuda.set_device(self.opt.gpu_ids[0])

        args = vars(self.opt)

        # #print('------------ Options -------------')
        # for k, v in sorted(args.items()):
        #     #print('%s: %s' % (str(k), str(v)))
        # #print('-------------- End ----------------')

        # save to the disk
        expr_dir = os.path.join(self.opt.checkpoints_dir, self.opt.name)
        util.mkdirs(expr_dir)
        file_name = os.path.join(expr_dir, 'opt.txt')
        with open(file_name, 'wt') as opt_file:
            opt_file.write('------------ Options -------------\n')
            for k, v in sorted(args.items()):
                opt_file.write('%s: %s\n' % (str(k), str(v)))
            opt_file.write('-------------- End ----------------\n')
            
        return self.opt


import cv2
import numpy as np
from matplotlib import pyplot as plt
from scipy.misc import imsave
from scipy import ndimage
from scipy import misc
import scipy.misc
import scipy
import image_slicer
from image_slicer import join
from PIL import Image


def name(self):
    return 'UnalignedDataset'



"""Model class template

This module provides a template for users to implement custom models.
You can specify '--model template' to use this model.
The class name should be consistent with both the filename and its model option.
The filename should be <model>_dataset.py
The class name should be <Model>Dataset.py
It implements a simple image-to-image translation baseline based on regression loss.
Given input-output pairs (data_A, data_B), it learns a network netG that can minimize the following L1 loss:
    min_<netG> ||netG(data_A) - data_B||_1
You need to implement the following functions:
    <modify_commandline_options>:　Add model-specific options and rewrite default values for existing options.
    <__init__>: Initialize this model class.
    <set_input>: Unpack input data and perform data pre-processing.
    <forward>: Run forward pass. This will be called by both <optimize_parameters> and <test>.
    <optimize_parameters>: Update network weights; it will be called in every training iteration.
"""
import torch
from .base_model import BaseModel
from . import networks


class TemplateModel(BaseModel):
    @staticmethod
    def modify_commandline_options(parser, is_train=True):
        """Add new model-specific options and rewrite default values for existing options.

        Parameters:
            parser -- the option parser
            is_train -- if it is training phase or test phase. You can use this flag to add training-specific or test-specific options.

        Returns:
            the modified parser.
        """
        parser.set_defaults(dataset_mode='aligned')  # You can rewrite default values for this model. For example, this model usually uses aligned dataset as its dataset.
        if is_train:
            parser.add_argument('--lambda_regression', type=float, default=1.0, help='weight for the regression loss')  # You can define new arguments for this model.

        return parser

    def __init__(self, opt):
        """Initialize this model class.

        Parameters:
            opt -- training/test options

        A few things can be done here.
        - (required) call the initialization function of BaseModel
        - define loss function, visualization images, model names, and optimizers
        """
        BaseModel.__init__(self, opt)  # call the initialization method of BaseModel
        # specify the training losses you want to print out. The program will call base_model.get_current_losses to plot the losses to the console and save them to the disk.
        self.loss_names = ['loss_G']
        # specify the images you want to save and display. The program will call base_model.get_current_visuals to save and display these images.
        self.visual_names = ['data_A', 'data_B', 'output']
        # specify the models you want to save to the disk. The program will call base_model.save_networks and base_model.load_networks to save and load networks.
        # you can use opt.isTrain to specify different behaviors for training and test. For example, some networks will not be used during test, and you don't need to load them.
        self.model_names = ['G']
        # define networks; you can use opt.isTrain to specify different behaviors for training and test.
        self.netG = networks.define_G(opt.input_nc, opt.output_nc, opt.ngf, opt.netG, gpu_ids=self.gpu_ids)
        if self.isTrain:  # only defined during training time
            # define your loss functions. You can use losses provided by torch.nn such as torch.nn.L1Loss.
            # We also provide a GANLoss class "networks.GANLoss". self.criterionGAN = networks.GANLoss().to(self.device)
            self.criterionLoss = torch.nn.L1Loss()
            # define and initialize optimizers. You can define one optimizer for each network.
            # If two networks are updated at the same time, you can use itertools.chain to group them. See cycle_gan_model.py for an example.
            self.optimizer = torch.optim.Adam(self.netG.parameters(), lr=opt.lr, betas=(opt.beta1, 0.999))
            self.optimizers = [self.optimizer]

        # Our program will automatically call <model.setup> to define schedulers, load networks, and print networks

    def set_input(self, input):
        """Unpack input data from the dataloader and perform necessary pre-processing steps.

        Parameters:
            input: a dictionary that contains the data itself and its metadata information.
        """
        AtoB = self.opt.direction == 'AtoB'  # use <direction> to swap data_A and data_B
        self.data_A = input['A' if AtoB else 'B'].to(self.device)  # get image data A
        self.data_B = input['B' if AtoB else 'A'].to(self.device)  # get image data B
        self.image_paths = input['A_paths' if AtoB else 'B_paths']  # get image paths

    def forward(self):
        """Run forward pass. This will be called by both functions <optimize_parameters> and <test>."""
        self.output = self.netG(self.data_A)  # generate output image given the input data_A

    def backward(self):
        """Calculate losses, gradients, and update network weights; called in every training iteration"""
        # caculate the intermediate results if necessary; here self.output has been computed during function <forward>
        # calculate loss given the input and intermediate results
        self.loss_G = self.criterionLoss(self.output, self.data_B) * self.opt.lambda_regression
        self.loss_G.backward()       # calculate gradients of network G w.r.t. loss_G

    def optimize_parameters(self):
        """Update network weights; it will be called in every training iteration."""
        self.forward()               # first call forward to calculate intermediate results
        self.optimizer.zero_grad()   # clear network G's existing gradients
        self.backward()              # calculate gradients for network G
        self.optimizer.step()        # update gradients for network G



import os.path
import random
import torchvision.transforms as transforms
import torch
from data.base_dataset import BaseDataset
from data.image_folder import make_dataset
from PIL import Image


class AlignedDataset(BaseDataset):
    def initialize(self, opt):
        self.opt = opt
        self.root = opt.dataroot
        self.dir_AB = os.path.join(opt.dataroot, opt.phase)
        self.AB_paths = sorted(make_dataset(self.dir_AB))
        assert(opt.resize_or_crop == 'resize_and_crop')

    def __getitem__(self, index):
        AB_path = self.AB_paths[index]
        AB = Image.open(AB_path).convert('RGB')
        AB = AB.resize((self.opt.loadSize * 2, self.opt.loadSize), Image.BICUBIC)
        AB = transforms.ToTensor()(AB)

        w_total = AB.size(2)
        w = int(w_total / 2)
        h = AB.size(1)
        w_offset = random.randint(0, max(0, w - self.opt.fineSize - 1))
        h_offset = random.randint(0, max(0, h - self.opt.fineSize - 1))

        A = AB[:, h_offset:h_offset + self.opt.fineSize,
               w_offset:w_offset + self.opt.fineSize]
        B = AB[:, h_offset:h_offset + self.opt.fineSize,
               w + w_offset:w + w_offset + self.opt.fineSize]

        A = transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))(A)
        B = transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))(B)

        if self.opt.which_direction == 'BtoA':
            input_nc = self.opt.output_nc
            output_nc = self.opt.input_nc
        else:
            input_nc = self.opt.input_nc
            output_nc = self.opt.output_nc

        if (not self.opt.no_flip) and random.random() < 0.5:
            idx = [i for i in range(A.size(2) - 1, -1, -1)]
            idx = torch.LongTensor(idx)
            A = A.index_select(2, idx)
            B = B.index_select(2, idx)

        if input_nc == 1:  # RGB to gray
            tmp = A[0, ...] * 0.299 + A[1, ...] * 0.587 + A[2, ...] * 0.114
            A = tmp.unsqueeze(0)

        if output_nc == 1:  # RGB to gray
            tmp = B[0, ...] * 0.299 + B[1, ...] * 0.587 + B[2, ...] * 0.114
            B = tmp.unsqueeze(0)

        return {'A': A, 'B': B,
                'A_paths': AB_path, 'B_paths': AB_path}

    def __len__(self):
        return len(self.AB_paths)

    def name(self):
        return 'AlignedDataset'



import time
import os, sys
from options.test_options import TestOptions
from data.data_loader import CreateDataLoader
from models.models import create_model
from util.visualizer_time import Visualizer

opt = TestOptions().parse()
opt.nThreads = 1  # test code only supports nThreads = 1
opt.batchSize = 1  # test code only supports batchSize = 1
opt.serial_batches = True  # no shuffle
opt.no_flip = True  # no flip

data_loader = CreateDataLoader(opt)
dataset = data_loader.load_data()
model = create_model(opt)
visualizer = Visualizer(opt)
# create website
# web_dir = os.path.join(opt.results_dir, opt.name, '%s_%s' % (opt.phase, opt.which_epoch))

web_dir = os.path.join(opt.dataroot, '_StainGAN')
if opt.results_dir:
    web_dir = opt.results_dir

# #print("web_dir ", web_dir)
# sys.exit()
#webpage = html.HTML(web_dir, 'Experiment = %s, Phase = %s, Epoch = %s' % (opt.name, opt.phase, opt.which_epoch))
# test
#print("Dataset", len(dataset))
start_time = time.time()

for i, data in enumerate(dataset):
    if opt.how_many:
        #print("how_many", opt.how_many)

        if i >= opt.how_many:
            break
    model.set_input(data)
    model.test()
    visuals = model.get_current_visuals()
    img_path = model.get_image_paths()
    # #print('%04d: process image... %s' % (i, img_path))
    #print('%04d: process image' % (i))
    visualizer.save_images(visuals, img_path, aspect_ratio=opt.aspect_ratio)

elapsed = (time.time() - start_time)
#print("--- %s seconds ---" % round((elapsed ), 2))

# webpage.save()




import numpy as np
from skimage import io
from skimage.color import rgb2hsv
from skimage.util import img_as_ubyte
from skimage import img_as_ubyte
from os import path, makedirs
from os.path import join 
import sys


openslide_error = False
try:
    import openslide as slide
except OSError as e:
    if "libopenslide.so.0" in str(e):
        openslide_error = True
    else:
        raise 


class SlidePatcher():
    
    def __init__(self,patch_size=224 , step_size=10, folder='./patch_output',patched_folder='./patched',thumbnails_folder='./thumbnails'):
        if openslide_error:
            print("An error occurred due to missing 'libopenslide.so.0'.")
            print("Please ensure that libopenslide is installed on your system.")
            print("For Debian/Ubuntu: sudo apt-get install libopenslide0")
            print("For Fedora/Red Hat: sudo dnf install openslide")
            print("For macOS (using Homebrew): brew install openslide")
            print("Refer to the official OpenSlide documentation for more detailed installation instructions.")

        self.patch_size = patch_size
        self.step_size = step_size
        self.folder = folder
        self.patched_folder = patched_folder
        self.thumbnails_folder = thumbnails_folder
        
    def check_margins(self,img, t=15):
        img = rgb2hsv(img)
        h, w, c = img.shape
        sat_img = img[:, :, 1]
        sat_img = img_as_ubyte(sat_img)
        ave_sat = np.sum(sat_img) / (h * w)
        return ave_sat >= t

    def crop_slide(self,img, save_slide_path, position=(0, 0), step=(0, 0), patch_size=224): # position given as (x, y) 
            img = img.read_region((position[0] * 4, position[1] * 4), 1, (patch_size, patch_size))
            img = np.array(img)[..., :3]
            if self.check_margins(img, 30):
                patch_name = "{}_{}".format(step[0], step[1])
                io.imsave(join(save_slide_path, patch_name + ".jpg"), img_as_ubyte(img))       
                            
    def patch(self,slides):
        print('Initializing patching process...')
        if type(slides) == str:
            slides = [slides]
        makedirs(self.folder, exist_ok=True)
        out_base = self.folder+self.patched_folder
        makedirs(out_base, exist_ok=True)
        makedirs(self.folder+self.thumbnails_folder, exist_ok=True)
        
        for s in range(len(slides)):
            img_slide = slides[s]
            img_name = img_slide.split(path.sep)[-1].split('.')[0]
            bag_path = join(out_base, img_name)
            makedirs(bag_path, exist_ok=True)
            img = slide.OpenSlide(img_slide)
            dimension = img.level_dimensions[1] # given as width, height
            if self.folder=='test':
                thumbnail = np.array(img.get_thumbnail((int(dimension[0])/7, int(dimension[1])/7)))[..., :3]
            else:
                thumbnail = np.array(img.get_thumbnail((int(dimension[0])/28, int(dimension[1])/28)))[..., :3]
            #io.imsave(join(self.folder, 'thumbnails', img_name + ".png"), img_as_ubyte(thumbnail))        
            step_y_max = int(np.floor(dimension[1]/self.step_size)) # rows
            step_x_max = int(np.floor(dimension[0]/self.step_size)) # columns
            for j in range(step_y_max): # rows
                for i in range(step_x_max): # columns
                    self.crop_slide(img, bag_path, (i*self.step_size, j*self.step_size), step=(j, i), patch_size=self.patch_size)
                sys.stdout.write('\r Cropped: {}/{} -- {}/{}'.format(s+1, len(slides), j+1, step_y_max))



import numpy as np
import os
import ntpath
import time
from . import util
from . import html


class Visualizer():
    def __init__(self, opt):
        # self.opt = opt
        self.display_id = opt.display_id
        self.use_html = opt.isTrain and not opt.no_html
        self.win_size = opt.display_winsize
        self.name = opt.name
        self.saved = False
        if self.display_id > 0:
            import visdom
            self.vis = visdom.Visdom(port=opt.display_port)
            self.display_single_pane_ncols = opt.display_single_pane_ncols

        if self.use_html:
            self.web_dir = os.path.join(opt.checkpoints_dir, opt.name, 'web')
            self.img_dir = os.path.join(self.web_dir, 'images')
            #print('create web directory %s...' % self.web_dir)
            util.mkdirs([self.web_dir, self.img_dir])
        self.log_name = os.path.join(opt.checkpoints_dir, opt.name, 'loss_log.txt')
        with open(self.log_name, "a") as log_file:
            now = time.strftime("%c")
            log_file.write('================ Training Loss (%s) ================\n' % now)

    def reset(self):
        self.saved = False

    # |visuals|: dictionary of images to display or save
    def display_current_results(self, visuals, epoch, save_result):
        if self.display_id > 0:  # show images in the browser
            if self.display_single_pane_ncols > 0:
                h, w = next(iter(visuals.values())).shape[:2]
                table_css = """<style>
                        table {border-collapse: separate; border-spacing:4px; white-space:nowrap; text-align:center}
                        table td {width: %dpx; height: %dpx; padding: 4px; outline: 4px solid black}
                        </style>""" % (w, h)
                ncols = self.display_single_pane_ncols
                title = self.name
                label_html = ''
                label_html_row = ''
                nrows = int(np.ceil(len(visuals.items()) / ncols))
                images = []
                idx = 0
                for label, image_numpy in visuals.items():
                    label_html_row += '<td>%s</td>' % label
                    images.append(image_numpy.transpose([2, 0, 1]))
                    idx += 1
                    if idx % ncols == 0:
                        label_html += '<tr>%s</tr>' % label_html_row
                        label_html_row = ''
                white_image = np.ones_like(image_numpy.transpose([2, 0, 1]))*255
                while idx % ncols != 0:
                    images.append(white_image)
                    label_html_row += '<td></td>'
                    idx += 1
                if label_html_row != '':
                    label_html += '<tr>%s</tr>' % label_html_row
                # pane col = image row
                self.vis.images(images, nrow=ncols, win=self.display_id + 1,
                                padding=2, opts=dict(title=title + ' images'))
                label_html = '<table>%s</table>' % label_html
                self.vis.text(table_css + label_html, win=self.display_id + 2,
                              opts=dict(title=title + ' labels'))
            else:
                idx = 1
                for label, image_numpy in visuals.items():
                    self.vis.image(image_numpy.transpose([2, 0, 1]), opts=dict(title=label),
                                   win=self.display_id + idx)
                    idx += 1

        if self.use_html and (save_result or not self.saved):  # save images to a html file
            self.saved  = True
            for label, image_numpy in visuals.items():
                img_path = os.path.join(self.img_dir, 'epoch%.3d_%s.png' % (epoch, label))
                util.save_image(image_numpy, img_path)
            # update website
            webpage = html.HTML(self.web_dir, 'Experiment name = %s' % self.name, reflesh=1)
            for n in range(epoch, 0, -1):
                webpage.add_header('epoch [%d]' % n)
                ims = []
                txts = []
                links = []

                for label, image_numpy in visuals.items():
                    img_path = 'epoch%.3d_%s.png' % (n, label)
                    ims.append(img_path)
                    txts.append(label)
                    links.append(img_path)
                webpage.add_images(ims, txts, links, width=self.win_size)
            webpage.save()

    # errors: dictionary of error labels and values
    def plot_current_errors(self, epoch, counter_ratio, opt, errors):
        if not hasattr(self, 'plot_data'):
            self.plot_data = {'X': [], 'Y': [], 'legend': list(errors.keys())}
        self.plot_data['X'].append(epoch + counter_ratio)
        self.plot_data['Y'].append([errors[k] for k in self.plot_data['legend']])
        self.vis.line(
            X=np.stack([np.array(self.plot_data['X'])] * len(self.plot_data['legend']), 1),
            Y=np.array(self.plot_data['Y']),
            opts={
                'title': self.name + ' loss over time',
                'legend': self.plot_data['legend'],
                'xlabel': 'epoch',
                'ylabel': 'loss'},
            win=self.display_id)

    # errors: same format as |errors| of plotCurrentErrors
    def #print_current_errors(self, epoch, i, errors, t):
        message = '(epoch: %d, iters: %d, time: %.3f) ' % (epoch, i, t)
        for k, v in errors.items():
            message += '%s: %.3f ' % (k, v)

        #print(message)
        with open(self.log_name, "a") as log_file:
            log_file.write('%s\n' % message)

    # save image to the disk
    def save_images(self, webpage, visuals, image_path):
        image_dir = webpage.get_image_dir()
        short_path = ntpath.basename(image_path[0])
        name = os.path.splitext(short_path)[0]

        webpage.add_header(name)
        ims = []
        txts = []
        links = []

        for label, image_numpy in visuals.items():
            image_name = '%s_%s.png' % (name, label)
            save_path = os.path.join(image_dir, image_name)
            util.save_image(image_numpy, save_path)

            ims.append(image_name)
            txts.append(label)
            links.append(image_name)
        webpage.add_images(ims, txts, links, width=self.win_size)




import numpy as np
import os
from PIL import Image


import os
import numpy as np
from PIL import Image

def LoadImages(input_data, return_filenames=False):
    if isinstance(input_data, np.ndarray):  
        return [input_data],None if return_filenames else [input_data],None
    elif isinstance(input_data, str):  
        if os.path.isfile(input_data):  
            image = Image.open(input_data)
            image_array = np.array(image).astype(np.float32)
            return [image_array], [os.path.basename(input_data)] if return_filenames else [image_array],[input_data]
        elif os.path.isdir(input_data):  
            images = []
            out_filenames = []
            out_path = []
            for filename in os.listdir(input_data):
                if filename.endswith(('.png', '.jpg', '.jpeg')):
                    image_path = os.path.join(input_data, filename)
                    out_filenames.append(os.path.basename(image_path))
                    out_path.append(image_path)
                    image = Image.open(image_path)
                    image_array = np.array(image).astype(np.float32)
                    if return_filenames:
                        images.append((image_array, filename))
                    else:
                        images.append(image_array)
                    
            return images, out_filenames,out_path
        else:
            raise ValueError("Invalid input path.")
    elif isinstance(input_data, Image.Image):  
        image_array = np.array(input_data).astype(np.float32)
        return [image_array], None if return_filenames else [image_array],None
    else:
        raise ValueError("Input must be a NumPy array, a string (path to an image or folder), or a Pillow Image.")

def SaveImages(images, output_dir, image_names=None):
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    saved_image_names = []  # List to store the names of the saved images

    for i, image in enumerate(images):
        # Check if a name is provided and valid, else generate a default name
        if image_names and i < len(image_names) and image_names[i]:
            filename = image_names[i]
        else:
            filename = f"image_{i}.jpg"

        # Ensure the filename has the correct extension
        filename_with_extension = os.path.splitext(filename)[0] + '.jpg'
        image_path = os.path.join(output_dir, filename_with_extension)

        # Convert and save the image
        pil_image = Image.fromarray(image.astype(np.uint8))
        pil_image.save(image_path)

        # Add the filename to the list of saved image names
        saved_image_names.append(filename_with_extension)

    return saved_image_names


import torch
import torch.nn as nn
from torch.nn import init
import functools
from torch.autograd import Variable
from torch.optim import lr_scheduler
import numpy as np


###############################################################################
# Functions
###############################################################################


def weights_init_normal(m):
    classname = m.__class__.__name__
    # #print(classname)
    if classname.find('Conv') != -1:
        init.normal(m.weight.data, 0.0, 0.02)
    elif classname.find('Linear') != -1:
        init.normal(m.weight.data, 0.0, 0.02)
    elif classname.find('BatchNorm2d') != -1:
        init.normal(m.weight.data, 1.0, 0.02)
        init.constant(m.bias.data, 0.0)


def weights_init_xavier(m):
    classname = m.__class__.__name__
    # #print(classname)
    if classname.find('Conv') != -1:
        init.xavier_normal(m.weight.data, gain=0.02)
    elif classname.find('Linear') != -1:
        init.xavier_normal(m.weight.data, gain=0.02)
    elif classname.find('BatchNorm2d') != -1:
        init.normal(m.weight.data, 1.0, 0.02)
        init.constant(m.bias.data, 0.0)


def weights_init_kaiming(m):
    classname = m.__class__.__name__
    # #print(classname)
    if classname.find('Conv') != -1:
        init.kaiming_normal(m.weight.data, a=0, mode='fan_in')
    elif classname.find('Linear') != -1:
        init.kaiming_normal(m.weight.data, a=0, mode='fan_in')
    elif classname.find('BatchNorm2d') != -1:
        init.normal(m.weight.data, 1.0, 0.02)
        init.constant(m.bias.data, 0.0)


def weights_init_orthogonal(m):
    classname = m.__class__.__name__
    #print(classname)
    if classname.find('Conv') != -1:
        init.orthogonal(m.weight.data, gain=1)
    elif classname.find('Linear') != -1:
        init.orthogonal(m.weight.data, gain=1)
    elif classname.find('BatchNorm2d') != -1:
        init.normal(m.weight.data, 1.0, 0.02)
        init.constant(m.bias.data, 0.0)


def init_weights(net, init_type='normal'):
    #print('initialization method [%s]' % init_type)
    if init_type == 'normal':
        net.apply(weights_init_normal)
    elif init_type == 'xavier':
        net.apply(weights_init_xavier)
    elif init_type == 'kaiming':
        net.apply(weights_init_kaiming)
    elif init_type == 'orthogonal':
        net.apply(weights_init_orthogonal)
    else:
        raise NotImplementedError('initialization method [%s] is not implemented' % init_type)


def get_norm_layer(norm_type='instance'):
    if norm_type == 'batch':
        norm_layer = functools.partial(nn.BatchNorm2d, affine=True)
    elif norm_type == 'instance':
        norm_layer = functools.partial(nn.InstanceNorm2d, affine=False)
    elif norm_type == 'none':
        norm_layer = None
    else:
        raise NotImplementedError('normalization layer [%s] is not found' % norm_type)
    return norm_layer


def get_scheduler(optimizer, opt):
    if opt.lr_policy == 'lambda':
        def lambda_rule(epoch):
            lr_l = 1.0 - max(0, epoch + 1 + opt.epoch_count - opt.niter) / float(opt.niter_decay + 1)
            return lr_l

        scheduler = lr_scheduler.LambdaLR(optimizer, lr_lambda=lambda_rule)
    elif opt.lr_policy == 'step':
        scheduler = lr_scheduler.StepLR(optimizer, step_size=opt.lr_decay_iters, gamma=0.1)
    elif opt.lr_policy == 'plateau':
        scheduler = lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', factor=0.2, threshold=0.01, patience=5)
    else:
        return NotImplementedError('learning rate policy [%s] is not implemented', opt.lr_policy)
    return scheduler


def define_G(input_nc, output_nc, ngf, which_model_netG, norm='batch', use_dropout=False, init_type='normal',
             gpu_ids=[]):
    netG = None
    use_gpu = len(gpu_ids) > 0
    norm_layer = get_norm_layer(norm_type=norm)

    # if use_gpu:
    #     assert (torch.cuda.is_available())

    if which_model_netG == 'resnet_9blocks':
        netG = ResnetGenerator(input_nc, output_nc, ngf, norm_layer=norm_layer, use_dropout=use_dropout, n_blocks=9,
                               gpu_ids=gpu_ids)
    elif which_model_netG == 'resnet_6blocks':
        netG = ResnetGenerator(input_nc, output_nc, ngf, norm_layer=norm_layer, use_dropout=use_dropout, n_blocks=6,
                               gpu_ids=gpu_ids)

    # densenet
    elif which_model_netG == 'densenet_6blocks':
        netG = ResnetGenerator(input_nc, output_nc, ngf, norm_layer=norm_layer, use_dropout=use_dropout, n_blocks=6,
                               gpu_ids=gpu_ids)




    elif which_model_netG == 'unet_128':
        netG = UnetGenerator(input_nc, output_nc, 7, ngf, norm_layer=norm_layer, use_dropout=use_dropout,
                             gpu_ids=gpu_ids)
    elif which_model_netG == 'unet_256':
        netG = UnetGenerator(input_nc, output_nc, 8, ngf, norm_layer=norm_layer, use_dropout=use_dropout,
                             gpu_ids=gpu_ids)
    else:
        raise NotImplementedError('Generator model name [%s] is not recognized' % which_model_netG)
    # if len(gpu_ids) > 0:
    #     netG.cuda(gpu_ids[0])
    init_weights(netG, init_type=init_type)
    return netG


def define_D(input_nc, ndf, which_model_netD,
             n_layers_D=3, norm='batch', use_sigmoid=False, init_type='normal', gpu_ids=[]):
    netD = None
    use_gpu = len(gpu_ids) > 0
    norm_layer = get_norm_layer(norm_type=norm)

    if use_gpu:
        assert (torch.cuda.is_available())
    if which_model_netD == 'basic':
        netD = NLayerDiscriminator(input_nc, ndf, n_layers=3, norm_layer=norm_layer, use_sigmoid=use_sigmoid,
                                   gpu_ids=gpu_ids)
    elif which_model_netD == 'n_layers':
        netD = NLayerDiscriminator(input_nc, ndf, n_layers_D, norm_layer=norm_layer, use_sigmoid=use_sigmoid,
                                   gpu_ids=gpu_ids)
    elif which_model_netD == 'pixel':
        netD = PixelDiscriminator(input_nc, ndf, norm_layer=norm_layer, use_sigmoid=use_sigmoid, gpu_ids=gpu_ids)
    else:
        raise NotImplementedError('Discriminator model name [%s] is not recognized' %
                                  which_model_netD)
    if use_gpu:
        netD.cuda(gpu_ids[0])
    init_weights(netD, init_type=init_type)
    return netD


def print_network(net):
    num_params = 0
    for param in net.parameters():
        num_params += param.numel()
    #print(net)
    #print('Total number of parameters: %d' % num_params)


##############################################################################
# Classes
##############################################################################


# Defines the GAN loss which uses either LSGAN or the regular GAN.  
# When LSGAN is used, it is basically same as MSELoss,
# but it abstracts away the need to create the target label tensor
# that has the same size as the input
class GANLoss(nn.Module):
    def __init__(self, use_lsgan=True, target_real_label=1.0, target_fake_label=0.0,
                 tensor=torch.FloatTensor):
        super(GANLoss, self).__init__()
        self.real_label = target_real_label
        self.fake_label = target_fake_label
        self.real_label_var = None
        self.fake_label_var = None
        self.Tensor = tensor
        if use_lsgan:
            self.loss = nn.MSELoss()
        else:
            self.loss = nn.BCELoss()

    def get_target_tensor(self, input, target_is_real):
        target_tensor = None
        if target_is_real:
            create_label = ((self.real_label_var is None) or
                            (self.real_label_var.numel() != input.numel()))
            if create_label:
                real_tensor = self.Tensor(input.size()).fill_(self.real_label)
                self.real_label_var = Variable(real_tensor, requires_grad=False)
            target_tensor = self.real_label_var
        else:
            create_label = ((self.fake_label_var is None) or
                            (self.fake_label_var.numel() != input.numel()))
            if create_label:
                fake_tensor = self.Tensor(input.size()).fill_(self.fake_label)
                self.fake_label_var = Variable(fake_tensor, requires_grad=False)
            target_tensor = self.fake_label_var
        return target_tensor

    def __call__(self, input, target_is_real):
        target_tensor = self.get_target_tensor(input, target_is_real)
        return self.loss(input, target_tensor)


# Defines the generator that consists of Resnet blocks between a few
# downsampling/upsampling operations.
# Code and idea originally from Justin Johnson's architecture.
# https://github.com/jcjohnson/fast-neural-style/
class ResnetGenerator(nn.Module):
    def __init__(self, input_nc, output_nc, ngf=64, norm_layer=nn.BatchNorm2d, use_dropout=False, n_blocks=6,
                 gpu_ids=[], padding_type='reflect'):
        assert (n_blocks >= 0)
        super(ResnetGenerator, self).__init__()
        self.input_nc = input_nc
        self.output_nc = output_nc
        self.ngf = ngf
        self.gpu_ids = gpu_ids
        if type(norm_layer) == functools.partial:
            use_bias = norm_layer.func == nn.InstanceNorm2d
        else:
            use_bias = norm_layer == nn.InstanceNorm2d

        model = [nn.ReflectionPad2d(3),
                 nn.Conv2d(input_nc, ngf, kernel_size=7, padding=0,
                           bias=use_bias),
                 norm_layer(ngf),
                 nn.ReLU(True)]

        n_downsampling = 2
        for i in range(n_downsampling):
            mult = 2 ** i
            model += [nn.Conv2d(ngf * mult, ngf * mult * 2, kernel_size=3,
                                stride=2, padding=1, bias=use_bias),
                      norm_layer(ngf * mult * 2),
                      nn.ReLU(True)]

        mult = 2 ** n_downsampling
        for i in range(n_blocks):
            model += [ResnetBlock(ngf * mult, padding_type=padding_type, norm_layer=norm_layer, use_dropout=use_dropout,
                                  use_bias=use_bias)]

        for i in range(n_downsampling):
            mult = 2 ** (n_downsampling - i)
            model += [nn.ConvTranspose2d(ngf * mult, int(ngf * mult / 2),
                                         kernel_size=3, stride=2,
                                         padding=1, output_padding=1,
                                         bias=use_bias),
                      norm_layer(int(ngf * mult / 2)),
                      nn.ReLU(True)]
        model += [nn.ReflectionPad2d(3)]
        model += [nn.Conv2d(ngf, output_nc, kernel_size=7, padding=0)]
        model += [nn.Tanh()]

        self.model = nn.Sequential(*model)

    def forward(self, input):
        if self.gpu_ids and isinstance(input.data, torch.cuda.FloatTensor):
            return nn.parallel.data_parallel(self.model, input, self.gpu_ids)
        else:
            return self.model(input)


# Define a resnet block
class ResnetBlock(nn.Module):
    def __init__(self, dim, padding_type, norm_layer, use_dropout, use_bias):
        super(ResnetBlock, self).__init__()
        self.conv_block = self.build_conv_block(dim, padding_type, norm_layer, use_dropout, use_bias)

    def build_conv_block(self, dim, padding_type, norm_layer, use_dropout, use_bias):
        conv_block = []
        p = 0
        if padding_type == 'reflect':
            conv_block += [nn.ReflectionPad2d(1)]
        elif padding_type == 'replicate':
            conv_block += [nn.ReplicationPad2d(1)]
        elif padding_type == 'zero':
            p = 1
        else:
            raise NotImplementedError('padding [%s] is not implemented' % padding_type)

        conv_block += [nn.Conv2d(dim, dim, kernel_size=3, padding=p, bias=use_bias),
                       norm_layer(dim),
                       nn.ReLU(True)]
        if use_dropout:
            conv_block += [nn.Dropout(0.5)]

        p = 0
        if padding_type == 'reflect':
            conv_block += [nn.ReflectionPad2d(1)]
        elif padding_type == 'replicate':
            conv_block += [nn.ReplicationPad2d(1)]
        elif padding_type == 'zero':
            p = 1
        else:
            raise NotImplementedError('padding [%s] is not implemented' % padding_type)
        conv_block += [nn.Conv2d(dim, dim, kernel_size=3, padding=p, bias=use_bias),
                       norm_layer(dim)]

        return nn.Sequential(*conv_block)

    def forward(self, x):
        out = x + self.conv_block(x)
        return out


## DENSENET
class DenseNetBlock(nn.Module):
    def __init__(self, in_channels=128, ks=3, padding=1, stride=1):
        super(DenseNetBlock, self).__init__()

        self.conv1 = nn.Sequential(
            nn.Conv2d(in_channels=in_channels, out_channels=16, kernel_size=3, stride=1, padding=1),
            nn.ReLU(inplace=True))

        self.conv2 = nn.Sequential(nn.Conv2d(16, 16, 3, 1, 1),
                                   nn.ReLU(inplace=True))

        self.conv3 = nn.Sequential(nn.Conv2d(32, 16, 3, 1, 1),
                                   nn.ReLU(inplace=True))

        self.conv4 = nn.Sequential(nn.Conv2d(48, 16, 3, 1, 1),
                                   nn.ReLU(inplace=True))

        self.conv5 = nn.Sequential(nn.Conv2d(64, 16, 3, 1, 1),
                                   nn.ReLU(inplace=True))

        self.conv6 = nn.Sequential(nn.Conv2d(80, 16, 3, 1, 1),
                                   nn.ReLU(inplace=True))

        self.conv7 = nn.Sequential(nn.Conv2d(96, 16, 3, 1, 1),
                                   nn.ReLU(inplace=True))

        self.conv8 = nn.Sequential(nn.Conv2d(112, 16, 3, 1, 1),
                                   nn.ReLU(inplace=True))

    def forward(self, x):
        # x = self.conv0(x) # Nx128x64x64
        conv1_out = self.conv1(x)  # Nx16x64x64

        conv2_out = self.conv2(conv1_out)  # Nx16x64x64

        conv3_in = torch.cat(([conv1_out, conv2_out]), 1)  # Nx32x64x64
        conv3_out = self.conv3(conv3_in)

        conv4_in = torch.cat(([conv1_out, conv2_out, conv3_out]), 1)  # Nx48x64x64
        conv4_out = self.conv4(conv4_in)

        conv5_in = torch.cat(([conv1_out, conv2_out, conv3_out, conv4_out]), 1)  # Nx64x64x64
        conv5_out = self.conv5(conv5_in)

        conv6_in = torch.cat(([conv1_out, conv2_out, conv3_out, conv4_out, conv5_out]), 1)  # Nx80x64x64
        conv6_out = self.conv6(conv6_in)

        conv7_in = torch.cat(([conv1_out, conv2_out, conv3_out, conv4_out, conv5_out, conv6_out]), 1)  # Nx96x64x64
        conv7_out = self.conv7(conv7_in)

        conv8_in = torch.cat(([conv1_out, conv2_out, conv3_out, conv4_out, conv5_out, conv6_out, conv7_out]),
                             1)  # Nx96x64x64
        conv8_out = self.conv8(conv8_in)

        output = torch.cat(([conv1_out, conv2_out, conv3_out, conv4_out,
                             conv5_out, conv6_out, conv7_out, conv8_out]), 1)

        return output


## DENSENET


# Defines the Unet generator.
# |num_downs|: number of downsamplings in UNet. For example,
# if |num_downs| == 7, image of size 128x128 will become of size 1x1
# at the bottleneck
class UnetGenerator(nn.Module):
    def __init__(self, input_nc, output_nc, num_downs, ngf=64,
                 norm_layer=nn.BatchNorm2d, use_dropout=False, gpu_ids=[]):
        super(UnetGenerator, self).__init__()
        self.gpu_ids = gpu_ids

        # construct unet structure
        unet_block = UnetSkipConnectionBlock(ngf * 8, ngf * 8, input_nc=None, submodule=None, norm_layer=norm_layer,
                                             innermost=True)
        for i in range(num_downs - 5):
            unet_block = UnetSkipConnectionBlock(ngf * 8, ngf * 8, input_nc=None, submodule=unet_block,
                                                 norm_layer=norm_layer, use_dropout=use_dropout)
        unet_block = UnetSkipConnectionBlock(ngf * 4, ngf * 8, input_nc=None, submodule=unet_block,
                                             norm_layer=norm_layer)
        unet_block = UnetSkipConnectionBlock(ngf * 2, ngf * 4, input_nc=None, submodule=unet_block,
                                             norm_layer=norm_layer)
        unet_block = UnetSkipConnectionBlock(ngf, ngf * 2, input_nc=None, submodule=unet_block, norm_layer=norm_layer)
        unet_block = UnetSkipConnectionBlock(output_nc, ngf, input_nc=input_nc, submodule=unet_block, outermost=True,
                                             norm_layer=norm_layer)

        self.model = unet_block

    def forward(self, input):
        if self.gpu_ids and isinstance(input.data, torch.cuda.FloatTensor):
            return nn.parallel.data_parallel(self.model, input, self.gpu_ids)
        else:
            return self.model(input)


# Defines the submodule with skip connection.
# X -------------------identity---------------------- X
#   |-- downsampling -- |submodule| -- upsampling --|
class UnetSkipConnectionBlock(nn.Module):
    def __init__(self, outer_nc, inner_nc, input_nc=None,
                 submodule=None, outermost=False, innermost=False, norm_layer=nn.BatchNorm2d, use_dropout=False):
        super(UnetSkipConnectionBlock, self).__init__()
        self.outermost = outermost
        if type(norm_layer) == functools.partial:
            use_bias = norm_layer.func == nn.InstanceNorm2d
        else:
            use_bias = norm_layer == nn.InstanceNorm2d
        if input_nc is None:
            input_nc = outer_nc
        downconv = nn.Conv2d(input_nc, inner_nc, kernel_size=4,
                             stride=2, padding=1, bias=use_bias)
        downrelu = nn.LeakyReLU(0.2, True)
        downnorm = norm_layer(inner_nc)
        uprelu = nn.ReLU(True)
        upnorm = norm_layer(outer_nc)

        if outermost:
            upconv = nn.ConvTranspose2d(inner_nc * 2, outer_nc,
                                        kernel_size=4, stride=2,
                                        padding=1)
            down = [downconv]
            up = [uprelu, upconv, nn.Tanh()]
            model = down + [submodule] + up
        elif innermost:
            upconv = nn.ConvTranspose2d(inner_nc, outer_nc,
                                        kernel_size=4, stride=2,
                                        padding=1, bias=use_bias)
            down = [downrelu, downconv]
            up = [uprelu, upconv, upnorm]
            model = down + up
        else:
            upconv = nn.ConvTranspose2d(inner_nc * 2, outer_nc,
                                        kernel_size=4, stride=2,
                                        padding=1, bias=use_bias)
            down = [downrelu, downconv, downnorm]
            up = [uprelu, upconv, upnorm]

            if use_dropout:
                model = down + [submodule] + up + [nn.Dropout(0.5)]
            else:
                model = down + [submodule] + up

        self.model = nn.Sequential(*model)

    def forward(self, x):
        if self.outermost:
            return self.model(x)
        else:
            return torch.cat([x, self.model(x)], 1)


# Defines the PatchGAN discriminator with the specified arguments.
class NLayerDiscriminator(nn.Module):
    def __init__(self, input_nc, ndf=64, n_layers=3, norm_layer=nn.BatchNorm2d, use_sigmoid=False, gpu_ids=[]):
        super(NLayerDiscriminator, self).__init__()
        self.gpu_ids = gpu_ids
        if type(norm_layer) == functools.partial:
            use_bias = norm_layer.func == nn.InstanceNorm2d
        else:
            use_bias = norm_layer == nn.InstanceNorm2d

        kw = 4
        padw = 1
        sequence = [
            nn.Conv2d(input_nc, ndf, kernel_size=kw, stride=2, padding=padw),
            nn.LeakyReLU(0.2, True)
        ]

        nf_mult = 1
        nf_mult_prev = 1
        for n in range(1, n_layers):
            nf_mult_prev = nf_mult
            nf_mult = min(2 ** n, 8)
            sequence += [
                nn.Conv2d(ndf * nf_mult_prev, ndf * nf_mult,
                          kernel_size=kw, stride=2, padding=padw, bias=use_bias),
                norm_layer(ndf * nf_mult),
                nn.LeakyReLU(0.2, True)
            ]

        nf_mult_prev = nf_mult
        nf_mult = min(2 ** n_layers, 8)
        sequence += [
            nn.Conv2d(ndf * nf_mult_prev, ndf * nf_mult,
                      kernel_size=kw, stride=1, padding=padw, bias=use_bias),
            norm_layer(ndf * nf_mult),
            nn.LeakyReLU(0.2, True)
        ]

        sequence += [nn.Conv2d(ndf * nf_mult, 1, kernel_size=kw, stride=1, padding=padw)]

        if use_sigmoid:
            sequence += [nn.Sigmoid()]

        self.model = nn.Sequential(*sequence)

    def forward(self, input):
        if len(self.gpu_ids) and isinstance(input.data, torch.cuda.FloatTensor):
            return nn.parallel.data_parallel(self.model, input, self.gpu_ids)
        else:
            return self.model(input)


class PixelDiscriminator(nn.Module):
    def __init__(self, input_nc, ndf=64, norm_layer=nn.BatchNorm2d, use_sigmoid=False, gpu_ids=[]):
        super(PixelDiscriminator, self).__init__()
        self.gpu_ids = gpu_ids
        if type(norm_layer) == functools.partial:
            use_bias = norm_layer.func == nn.InstanceNorm2d
        else:
            use_bias = norm_layer == nn.InstanceNorm2d

        self.net = [
            nn.Conv2d(input_nc, ndf, kernel_size=1, stride=1, padding=0),
            nn.LeakyReLU(0.2, True),
            nn.Conv2d(ndf, ndf * 2, kernel_size=1, stride=1, padding=0, bias=use_bias),
            norm_layer(ndf * 2),
            nn.LeakyReLU(0.2, True),
            nn.Conv2d(ndf * 2, 1, kernel_size=1, stride=1, padding=0, bias=use_bias)]

        if use_sigmoid:
            self.net.append(nn.Sigmoid())

        self.net = nn.Sequential(*self.net)

    def forward(self, input):
        if len(self.gpu_ids) and isinstance(input.data, torch.cuda.FloatTensor):
            return nn.parallel.data_parallel(self.net, input, self.gpu_ids)
        else:
            return self.net(input)



#%%
from stainnet.stain_net import Stainnet
from PIL import Image 
import matplotlib.pyplot as plt


# mode = ["histopathology", "camelyon", "cytopathology"]
checkpoint_dir ="./StainNet/checkpoints/aligned_cytopathology_dataset/StainNet-3x0_best_psnr_layer3_ch32.pth"
# img_source=Image.open("./StainNet/assets/3_color_net_neg23570_ori.png")
img_source=Image.open("./test.png")

# plt.imshow(img_source)

model = Stainnet(checkpoint_dir)
normalized_image = model.transform(img_source)
normalized_image = Image.fromarray(normalized_image)

plt.imshow(normalized_image)
plt.show()


import numpy as np
import logging
import os

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
logging.getLogger('tensorflow').setLevel(logging.ERROR)

# import tensorflow as tf
import tensorflow.compat.v1 as tf
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
tf.disable_v2_behavior()

# initial varphi for rgb input
init_varphi = np.asarray([[0.294, 0.110, 0.894],
                          [0.750, 0.088, 0.425]])

# # initial varphi for bgr input
# init_varphi = np.asarray([[0.6060, 1.2680, 0.7989],
#                           [1.2383, 1.2540, 0.3927]])

class AdaptiveColorDeconvolution():
    
    def __init__(self, pixel_number=10000, step=300, batch_size=1500):
        super().__init__()
        self._pn = pixel_number
        self._bs = batch_size
        self._step_per_epoch = int(pixel_number / batch_size)
        self._epoch = int(step / self._step_per_epoch)
        self._template_dc_mat = None
        self._template_w_mat = None
        
    
    def fit(self, images):
        opt_cd_mat, opt_w_mat = self.extract_adaptive_cd_params(images)
        self._template_dc_mat = opt_cd_mat
        self._template_w_mat = opt_w_mat
        
    
    def transform(self, images):
        if self._template_dc_mat is None:
            raise AssertionError('Run fit function first.')

        opt_cd_mat, opt_w_mat = self.extract_adaptive_cd_params(images)
        transform_mat = np.matmul(opt_cd_mat * opt_w_mat,
                                  np.linalg.inv(self._template_dc_mat * self._template_w_mat))

        od = -np.log((np.asarray(images, float) + 1) / 256.0)
        normed_od = np.matmul(od, transform_mat)
        normed_images = np.exp(-normed_od) * 256 - 1

        return np.maximum(np.minimum(normed_images, 255), 0)
    
    def he_decomposition(self, images, od_output=True):
        if self._template_dc_mat is None:
            raise AssertionError('Run fit function first.')

        opt_cd_mat, _ = self.extract_adaptive_cd_params(images)

        od = -np.log((np.asarray(images, float) + 1) / 256.0)
        normed_od = np.matmul(od, opt_cd_mat)

        if od_output:
            return normed_od
        else:
            normed_images = np.exp(-normed_od) * 256 - 1
            return np.maximum(np.minimum(normed_images, 255), 0)

    def sampling_data(self, images):
        pixels = np.reshape(images, (-1, 3))
        pixels = pixels[np.random.choice(pixels.shape[0], min(self._pn * 20, pixels.shape[0]))]
        od = -np.log((np.asarray(pixels, float) + 1) / 256.0)
        
        # filter the background pixels (white or black)
        tmp = np.mean(od, axis=1)
        od = od[(tmp > 0.3) & (tmp < -np.log(30 / 256))]
        od = od[np.random.choice(od.shape[0], min(self._pn, od.shape[0]))]

        return od

    def extract_adaptive_cd_params(self, images):
        """
        :param images: RGB uint8 format in shape of [k, m, n, 3], where
                       k is the number of ROIs sampled from a WSI, [m, n] is 
                       the size of ROI.
        """
        od_data = self.sampling_data(images)
        input_od = tf.placeholder(dtype=tf.float32, shape=[None, 3])
        target, cd, w = self.acd_model(input_od)
        init = tf.global_variables_initializer()

        with tf.Session() as sess:
            sess.run(init)
            for ep in range(self._epoch):
                for step in range(self._step_per_epoch):
                    sess.run(target, {input_od: od_data[step * self._bs:(step + 1) * self._bs]})
            opt_cd = sess.run(cd)
            opt_w = sess.run(w)
        return opt_cd, opt_w
    
    
    def acd_model(self, input_od, lambda_p=0.002, lambda_b=10, lambda_e=1, eta=0.6, gamma=0.5):
        """
        Stain matrix estimation by
        "Yushan Zheng, et al., Adaptive Color Deconvolution for Histological WSI Normalization."

        """
        alpha = tf.Variable(init_varphi[0], dtype='float32')
        beta = tf.Variable(init_varphi[1], dtype='float32')
        w = [tf.Variable(1.0, dtype='float32'), tf.Variable(1.0, dtype='float32'), tf.constant(1.0)]

        sca_mat = tf.stack((tf.cos(alpha) * tf.sin(beta), tf.cos(alpha) * tf.cos(beta), tf.sin(alpha)), axis=1)
        cd_mat = tf.matrix_inverse(sca_mat)

        s = tf.matmul(input_od, cd_mat) * w
        h, e, b = tf.split(s, (1, 1, 1), axis=1)

        l_p1 = tf.reduce_mean(tf.square(b))
        l_p2 = tf.reduce_mean(2 * h * e / (tf.square(h) + tf.square(e)))
        l_b = tf.square((1 - eta) * tf.reduce_mean(h) - eta * tf.reduce_mean(e))
        l_e = tf.square(gamma - tf.reduce_mean(s))

        objective = l_p1 + lambda_p * l_p2 + lambda_b * l_b + lambda_e * l_e
        target = tf.train.AdagradOptimizer(learning_rate=0.05).minimize(objective)

        return target, cd_mat, w


import random
import numpy as np
import torch
from torch.autograd import Variable


class ImagePool():
    def __init__(self, pool_size):
        self.pool_size = pool_size
        if self.pool_size > 0:
            self.num_imgs = 0
            self.images = []

    def query(self, images):
        if self.pool_size == 0:
            return Variable(images)
        return_images = []
        for image in images:
            image = torch.unsqueeze(image, 0)
            if self.num_imgs < self.pool_size:
                self.num_imgs = self.num_imgs + 1
                self.images.append(image)
                return_images.append(image)
            else:
                p = random.uniform(0, 1)
                if p > 0.5:
                    random_id = random.randint(0, self.pool_size-1)
                    tmp = self.images[random_id].clone()
                    self.images[random_id] = image
                    return_images.append(tmp)
                else:
                    return_images.append(image)
        return_images = Variable(torch.cat(return_images, 0))
        return return_images



import os
import cv2
import numpy as np
import time
import argparse
import yaml
import json
import random
import copy
from skimage import color
from fitter import Fitter

parser = argparse.ArgumentParser(description="norm&jitter dataset lab statistics")
parser.add_argument(
    "--data-dir", type=str, metavar="DIR", help="path to dataset(ImageFolder form)"
)
parser.add_argument("--save-dir", type=str, metavar="DIR", help="path to save dataset")
parser.add_argument(
    "--dataset-name", type=str, default="", metavar="DIR", help="dataset output_name"
)
parser.add_argument("--methods", type=str, default="", help="colornorm_methods")
parser.add_argument(
    "--color-space",
    type=str,
    default="LAB",
    choices=["LAB", "HED", "HSV"],
    help="dataset statistics color space",
)
parser.add_argument(
    "--random", action="store_true", default=False, help="random shuffle sample"
)
parser.add_argument(
    "--n",
    type=int,
    default=0,
    metavar="DIR",
    help="datasets statistics sample n image each class(0:all)",
)


def _parse_args():
    args = parser.parse_args()

    return args


def getavgstd(image):
    avg = []
    std = []
    image_avg_l = np.mean(image[:, :, 0])
    image_std_l = np.std(image[:, :, 0])
    image_avg_a = np.mean(image[:, :, 1])
    image_std_a = np.std(image[:, :, 1])
    image_avg_b = np.mean(image[:, :, 2])
    image_std_b = np.std(image[:, :, 2])
    avg.append(image_avg_l)
    avg.append(image_avg_a)
    avg.append(image_avg_b)
    std.append(image_std_l)
    std.append(image_std_a)
    std.append(image_std_b)
    return (avg, std)


if __name__ == "__main__":

    args = _parse_args()

    path_dataset = args.data_dir

    labL_avg_List = []
    labA_avg_List = []
    labB_avg_List = []
    labL_std_List = []
    labA_std_List = []
    labB_std_List = []

    t1 = time.time()
    i = 0

    for class_dir in os.listdir(path_dataset):
        path_class = os.path.join(path_dataset, class_dir)
        print(path_class)

        path_class_list = os.listdir(path_class)
        if args.random:
            random.shuffle(path_class_list)

        for image in path_class_list:
            if args.n == 0:  # n=0: all images each class
                pass
            elif i < args.n:
                i += 1
            else:
                i = 0
                break
            path_img = os.path.join(path_class, image)
            img = cv2.imread(path_img)
            try:  # debug
                if args.color_space == "LAB":
                    img = cv2.cvtColor(img, cv2.COLOR_BGR2LAB)
                elif args.color_space == "HED":
                    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                    img = color.rgb2hed(img)
                elif args.color_space == "HSV":
                    img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
                else:
                    print("wrong color space: {}!!".format(args.color_space))
                img_avg, img_std = getavgstd(img)
            except:
                continue
                print(path_img)
            labL_avg_List.append(img_avg[0])
            labA_avg_List.append(img_avg[1])
            labB_avg_List.append(img_avg[2])
            labL_std_List.append(img_std[0])
            labA_std_List.append(img_std[1])
            labB_std_List.append(img_std[2])
    t2 = time.time()
    print(t2 - t1)
    l_avg_mean = np.mean(labL_avg_List).item()
    l_avg_std = np.std(labL_avg_List).item()
    l_std_mean = np.mean(labL_std_List).item()
    l_std_std = np.std(labL_std_List).item()
    a_avg_mean = np.mean(labA_avg_List).item()
    a_avg_std = np.std(labA_avg_List).item()
    a_std_mean = np.mean(labA_std_List).item()
    a_std_std = np.std(labA_std_List).item()
    b_avg_mean = np.mean(labB_avg_List).item()
    b_avg_std = np.std(labB_avg_List).item()
    b_std_mean = np.mean(labB_std_List).item()
    b_std_std = np.std(labB_std_List).item()

    std_avg_list = [
        labL_avg_List,
        labL_std_List,
        labA_avg_List,
        labA_std_List,
        labB_avg_List,
        labB_std_List,
    ]
    distribution = []
    for std_avg in std_avg_list:
        f = Fitter(std_avg, distributions=["norm", "laplace"])
        f.fit()
        distribution.append(list(f.get_best(method="sumsquare_error").keys())[0])

    yaml_dict_lab = {
        "random": args.random,
        "n_each_class": args.n,
        "color_space": args.color_space,
        "methods": args.methods,
        "{}".format(args.color_space[0]): {  # lab-L/hed-H
            "avg": {
                "mean": round(l_avg_mean, 3),
                "std": round(l_avg_std, 3),
                "distribution": distribution[0],
            },
            "std": {
                "mean": round(l_std_mean, 3),
                "std": round(l_std_std, 3),
                "distribution": distribution[1],
            },
        },
        "{}".format(args.color_space[1]): {  # lab-A/hed-E
            "avg": {
                "mean": round(a_avg_mean, 3),
                "std": round(a_avg_std, 3),
                "distribution": distribution[2],
            },
            "std": {
                "mean": round(a_std_mean, 3),
                "std": round(a_std_std, 3),
                "distribution": distribution[3],
            },
        },
        "{}".format(args.color_space[2]): {  # lab-B/hed-D
            "avg": {
                "mean": round(b_avg_mean, 3),
                "std": round(b_avg_std, 3),
                "distribution": distribution[4],
            },
            "std": {
                "mean": round(b_std_mean, 3),
                "std": round(b_std_std, 3),
                "distribution": distribution[5],
            },
        },
    }
    yaml_save_path = "{}/{}.yaml".format(
        args.save_dir,
        args.dataset_name
        if args.dataset_name != ""
        else "dataset_{}_random{}_n{}".format(args.color_space, args.random, args.n),
    )
    with open(yaml_save_path, "w") as f:
        yaml.dump(yaml_dict_lab, f)
        print("The dataset lab statistics has been saved in {}".format(yaml_save_path))


import torch.utils.data
from ..data.base_data_loader import BaseDataLoader


def CreateDataset(opt , path_image):
    dataset = None
    if opt.dataset_mode == 'aligned':
        from ..data.aligned_dataset import AlignedDataset
        dataset = AlignedDataset()
    elif opt.dataset_mode == 'unaligned':
        from ..data.unaligned_dataset import UnalignedDataset
        dataset = UnalignedDataset()
    elif opt.dataset_mode == 'single':
        from ..data.single_dataset import SingleDataset
        dataset = SingleDataset()
    else:
        raise ValueError("Dataset [%s] not recognized." % opt.dataset_mode)

    #print("dataset [%s] was created" % (dataset.name()))
    dataset.initialize(opt, path_image = path_image)
    return dataset


class CustomDatasetDataLoader(BaseDataLoader):
    # self.path_image = path_image
    def name(self):
        return 'CustomDatasetDataLoader'

    def initialize(self, opt , path_image):
        BaseDataLoader.initialize(self, opt)
        self.dataset = CreateDataset(opt,path_image = path_image)
        self.dataloader = torch.utils.data.DataLoader(
            self.dataset,
            batch_size=opt.batchSize,
            shuffle=not opt.serial_batches,
            num_workers=int(opt.nThreads))

    def load_data(self):
        return self

    def __len__(self):
        return min(len(self.dataset), self.opt.max_dataset_size)

    def __iter__(self):
        for i, data in enumerate(self.dataloader):
            if i >= self.opt.max_dataset_size:
                break
            yield data



import time, sys
from options.train_options import TrainOptions
from data.data_loader import CreateDataLoader
from models.models import create_model
from util.visualizer_time import Visualizer

opt = TrainOptions().parse()
data_loader = CreateDataLoader(opt)
dataset = data_loader.load_data()
dataset_size = len(data_loader)
#print('#training images = %d' % dataset_size)

model = create_model(opt)
visualizer = Visualizer(opt)
total_steps = 0

# for epoch in range(opt.epoch_count, opt.niter + opt.niter_decay + 1):

train_start_time = time.time()
no_saved = 1
frequency_count = 1

# REAL MODEL
train_end_time = train_start_time + (3600 * 4)
save_model_frequency = 3600
display_frequency = 300
#print_frequency = 300

# TRAIL MODEL
# train_end_time = train_start_time + (180 * 4)
# save_model_frequency = 180
# display_frequency = 60
# #print_frequency = 60
#print(len(dataset))
while True:

    # #print(epoch_start_time)
    # sys.exit()
    # epoch_iter = 0

    for i, data in enumerate(dataset):
        #print("IN ITERATION ", i)
        iter_start_time = time.time()
        visualizer.reset()
        total_steps += opt.batchSize
        model.set_input(data)
        model.optimize_parameters()

        if time.time() >= train_end_time:
            model.save('latest-for-real')
            total_train_time = (train_end_time - train_start_time) / 3600
            #print("End of Training  start time %s, end time %s - total time %s" % (
                train_start_time, train_end_time, total_train_time))
            #print("EXITING TRAINING")
            sys.exit()
            break

        if time.time() >= train_start_time + (no_saved * save_model_frequency):
            #print('saving the model at the end of no_saved %d, time %s' %
                  (no_saved, time.time()))
            model.save(no_saved)
            no_saved = no_saved + 1

        if time.time() >= train_start_time + (frequency_count * display_frequency):
            frequency_count = frequency_count + 1
            visualizer.display_current_results(model.get_current_visuals(), no_saved, True)
            errors = model.get_current_errors()
            t = (time.time() - iter_start_time) / opt.batchSize
            visualizer.#print_current_errors(no_saved, frequency_count, errors, t)
            if opt.display_id > 0:
                visualizer.plot_current_errors(no_saved, float(frequency_count), opt, errors)

    if time.time() >= train_end_time:
        #print("EXITING TRAINING")
        sys.exit()
        break
    model.update_learning_rate()






from typing_extensions import Self
import statistics
from torchmetrics import MultiScaleStructuralSimilarityIndexMeasure
import torch
import lpips
import numpy
from image_similarity_measures.quality_metrics import fsim,ssim,psnr

class NMI:
  """
  Args:
    numpy array of images in N*W*H*C
  Example:
    myNMI=NMI(np.asarray([image1,image2])) # each image has W*H*C shape
    print(myNMI.NMIValue)
    print(myNMI.Mean)
    print(myNMI.Sd) 
    """
  def __init__(self,ImageList):
    self.ImageList=ImageList
    self.NMIValue=self.ComputeNMI()
    self.Mean=self.MeanNMI()
    self.Sd=self.SdNMI()

  def SingleNMI(Img):
    ImgPixelWiseMean=numpy.mean(Img, axis=2)
    ImgNMI=np.median(ImgPixelWiseMean)/np.max(ImgPixelWiseMean)
    return(ImgNMI)

  def ComputeNMI(self):
    return(list(map(NMI.SingleNMI,self.ImageList)))

  def MeanNMI(self):
    return(statistics.mean(self.NMIValue))

  def SdNMI(self):
    return(statistics.stdev(self.NMIValue))


class MSSSIM:
  """
  Args:
     2 numpy array of images in (B,W,H,C) or torch array with (B,C,W,H) shape
  Example:
      MSSSIM(images1,images2).measure
    """
  def __init__(self,ImageList1,ImageList2):
    if(type(ImageList1)==numpy.ndarray):
      self.ImageList1=ReorderChannels(ImageList1).ConvertedImages
      self.ImageList2=ReorderChannels(ImageList2).ConvertedImages
    if(type(ImageList1)==torch.Tensor):
      self.ImageList1=ImageList1
      self.ImageList2=ImageList2
    self.measure=self.ComputeMSSSIM()

  def ComputeMSSSIM(self):
    ms_ssim= MultiScaleStructuralSimilarityIndexMeasure()
    ms_ssim=ms_ssim(self.ImageList1,self.ImageList2)
    return(ms_ssim)


class ReorderChannels:
  def __init__(self,ImageList,OnlyNumpyToTorch=False):
    self.ImageList=ImageList
    self.ReorderedImages=self.Reorder()
    if(OnlyNumpyToTorch==True):
      self.ReorderedImages=ImageList
    self.ConvertedImages=self.NumpyToTorch()

  def Reorder(self):
    return(self.ImageList.transpose(0,3,1,2))
  def NumpyToTorch(self):
    return(torch.Tensor(self.ReorderedImages))


class LPIPSimilarity:
  """
  images should be RGB, IMPORTANT: normalized to [-1,1]
  set net as alex or vgg
  image should be normalized to [-1,1]
  Example:
    LPIPS=LPIPSimilarity()
    LPIPS.Computesimilarity(images1,images2)
    """
  def __init__(self,net="alex"):
    self.net=net
    self.loss_fn=lpips.LPIPS(net=self.net)

  def Computesimilarity(self,ImageList1,ImageList2):
    if(type(ImageList1)==numpy.ndarray):
      ImageList1=ReorderChannels(ImageList1,OnlyNumpyToTorch=True).ConvertedImages
      ImageList2=ReorderChannels(ImageList2,OnlyNumpyToTorch=True).ConvertedImages
    if(type(ImageList1)==torch.Tensor):
      ImageList1=ImageList1
      ImageList2=ImageList2
      similarity=1-self.loss_fn(ImageList1, ImageList2)
    return(similarity)

class FSIM:
  """
  Args:
      two numpy array with shape (N,W,H,C)
  Example:
      FSIM(images1,images2).measure
  """
  def __init__(self,ImageList1,ImageList2):
    self.ImageList1=ImageList1
    self.ImageList2=ImageList2
    self.measure=self.ComputeSimilarity()
  def ComputeSimilarity(self):
    return(list(map(fsim,self.ImageList1, self.ImageList2)))

class SSIM:
  """
  Args:
      two numpy array with shape (N,W,H,C)
  Example:
      FSIM(images1,images2).measure
  """
  def __init__(self,ImageList1,ImageList2):
    self.ImageList1=ImageList1
    self.ImageList2=ImageList2
    self.measure=self.ComputeSimilarity()
  def ComputeSimilarity(self):
    return(list(map(ssim,self.ImageList1, self.ImageList2)))

class PSNR:
  """
  Args:
      two numpy array with shape (N,W,H,C)
  Example:
      FSIM(images1,images2).measure
  """
  def __init__(self,ImageList1,ImageList2):
    self.ImageList1=ImageList1
    self.ImageList2=ImageList2
    self.measure=self.ComputeSimilarity()
  def ComputeSimilarity(self):
    return(list(map(psnr,self.ImageList1, self.ImageList2)))

class RMSE:
  """
  Args:
      two numpy array with shape (N,W,H,C)
  Example:
      FSIM(images1,images2).measure
  """
  def __init__(self,ImageList1,ImageList2):
    self.ImageList1=ImageList1
    self.ImageList2=ImageList2
    self.measure=self.ComputeSimilarity()
  def ComputeSimilarity(self):
    return(list(map(rmse,self.ImageList1, self.ImageList2)))




import time
import os, sys
from .options.test_options import TestOptions
from .data.data_loader import CreateDataLoader
from .models.models import create_model
from .util.visualizer_time import Visualizer
from .util import html
import requests
from tqdm import tqdm
import os
from ..image_handler import LoadImages

class StainGAN:
    
    
    def __init__(self, model_path=''):
        self.model_name = "staingan_model.pth"
        self.is_model_ready = model_path != ''
        self.model_path = model_path
        
        self._check_model(self._get_safe_path(self.model_name))
        
    def _get_safe_path(self,filename):
        home_dir = os.path.expanduser("~")
        dir_name = ".lbbnorm"
        safe_path = os.path.join(home_dir, dir_name)
        if not os.path.exists(safe_path):
            os.makedirs(safe_path)
        return os.path.join(safe_path, filename)
    
    def _check_model(self,model_path):
        
        if os.path.exists(model_path):
            self.model_path = model_path
            self.is_model_ready = True
            return True
        else:
            return False
    

    def download_model(self):
        public_url = "https://drive.google.com/file/d/1VzZeeUWDLpolY1_zECFD8MwSkbF1Ps37/view?usp=sharing"
        destination = self._get_safe_path(self.model_name)
        
        if self._check_model(destination):
            print('Model already exists.')
            return
        
        file_id = public_url.split('/')[-2]
        download_url = f"https://drive.google.com/uc?export=download&id={file_id}"
        response = requests.get(download_url, stream=True)
        total_size_in_bytes= int(response.headers.get('content-length', 0))
        block_size = 1024
        progress_bar = tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)
        with open(destination, 'wb') as file:
            for data in response.iter_content(block_size):
                progress_bar.update(len(data))
                file.write(data)
        progress_bar.close()
        if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
            print("ERROR, something went wrong, check your internet connection.")
            try:
                os.remove(destination)
            except:
                pass
        else:
            self.model_path = destination
            self.is_model_ready = True
  
    
    def transform(self, path_image, image_size,results_dir='./staingan_resluts/'):
        
        if not self._check_model(self.model_path) or len(self.model_path) < 1:
            print("""Pretrained model file not found, please download or specify it first.
Hint: you can download the pretrained model with `model.download_model()`, or you can specify it like this `StainGAN(model_path='../../../model.pth')` on creation time.""")
            FileNotFoundError()
            return
        
        _,_,path_image_list = LoadImages(path_image)
        
        print(path_image_list)
        
        output = []
        
        for path_image in tqdm(path_image_list):  
            opt = TestOptions().parse(image_size,results_dir)
            opt.nThreads = 1  # test code only supports nThreads = 1
            opt.batchSize = 1  # test code only supports batchSize = 1
            opt.serial_batches = True  # no shuffle
            opt.no_flip = True  # no flip
            
            data_loader = CreateDataLoader(opt,path_image)
            dataset = data_loader.load_data()
            model = create_model(opt,self.model_path)
            visualizer = Visualizer(opt)
            # create website
            # web_dir = os.path.join(opt.results_dir, opt.name, '%s_%s' % (opt.phase, opt.which_epoch))
            
            web_dir = os.path.join(opt.dataroot, '_StainGAN')
            if opt.results_dir:
                web_dir = opt.results_dir
            
            # #print("web_dir ", web_dir)
            # sys.exit()
            webpage = html.HTML(web_dir, 'Experiment = %s, Phase = %s, Epoch = %s' % (opt.name, opt.phase, opt.which_epoch))
            # test
            ##print("Dataset", len(dataset))
            start_time = time.time()
            
            for i, data in enumerate(dataset):
                if opt.how_many:
                    ##print("how_many", opt.how_many)
            
                    if i >= opt.how_many:
                        break
                model.set_input(data)
                model.test()
                visuals = model.get_current_visuals()
                img_path = model.get_image_paths()
                # #print('%04d: process image... %s' % (i, img_path))
                ##print('%04d: process image' % (i))
                img = visualizer.save_images(visuals, img_path, aspect_ratio=opt.aspect_ratio)
            
            output.append(img)
        
        if len(output)==1:
            return output[0]
        else:
            return output
        


from torch.autograd import Variable
from collections import OrderedDict
from ..util import util
from .base_model import BaseModel
from . import networks


class TestModel(BaseModel):
    def name(self):
        return 'TestModel'

    def initialize(self, opt,load_path):
        assert(not opt.isTrain)
        BaseModel.initialize(self, opt)
        self.netG = networks.define_G(opt.input_nc, opt.output_nc,
                                      opt.ngf, opt.which_model_netG,
                                      opt.norm, not opt.no_dropout,
                                      opt.init_type,
                                      self.gpu_ids)
        which_epoch = opt.which_epoch
        self.load_network(self.netG, 'G', which_epoch,load_path)

    def set_input(self, input):
        # we need to use single_dataset mode
        input_A = input['A']
        # if len(self.gpu_ids) > 0:
        #     input_A = input_A.cuda(self.gpu_ids[0], async=True)
        self.input_A = input_A
        self.image_paths = input['A_paths']

    def test(self):
        self.real_A = Variable(self.input_A)
        self.fake_B = self.netG(self.real_A)

    # get image paths
    def get_image_paths(self):
        return self.image_paths

    def get_current_visuals(self):
        real_A = util.tensor2im(self.real_A.data)
        fake_B = util.tensor2im(self.fake_B.data)
        return OrderedDict([('real_A', real_A), ('fake_B', fake_B)])



import os
from randstainna import RandStainNA
import cv2

if __name__ == "__main__":

    """
    Usage1: Demo(for visualization)
    """
    # Setting: is_train = False
    randstainna = RandStainNA(
        yaml_file = '/home/thecapitanprice/driveA/python/LLB/LBBNorm/src/LBBNorm/randstain/randstainna/CRC.yaml',
        std_hyper = 0.0,
        distribution = 'normal',
        probability = 1.0,
        is_train = False
    )

    img_path_list = os.listdir('./data/')
    save_dir_path = './visualization'
    if not os.path.exists(save_dir_path):
        os.mkdir(save_dir_path)

    for img_path in img_path_list:
        img = randstainna(cv2.imread('./data/'+img_path))
        save_img_path = save_dir_path + '/{}'.format(img_path.split('/')[-1])
        cv2.imwrite(save_img_path,img)

    """
    Usage2：torchvision.transforms (for training)
    """
    # # Setting: is_train = True
    # from torchvision import transforms

    # #### calling the randstainna
    # transforms_list = [
    #     RandStainNA(
    #         yaml_file="/home/thecapitanprice/driveA/python/LLB/LBBNorm/src/LBBNorm/randstain/randstainna/CRC.yaml",
    #         std_hyper=-0.3,
    #         probability=1.0,
    #         distribution="normal",
    #         is_train=True,
    #     )
    # ]
    # transforms.Compose(transforms_list)


from .base_options import BaseOptions


class TestOptions(BaseOptions):
    def initialize(self, path_size,results_dir):
        BaseOptions.initialize(self, path_size=path_size)
        self.parser.add_argument('--ntest', type=int, default=float("inf"), help='# of test examples.')
        self.parser.add_argument('--results_dir', type=str, default=results_dir, help='saves results here.')
        self.parser.add_argument('--aspect_ratio', type=float, default=1.0, help='aspect ratio of result images')
        self.parser.add_argument('--phase', type=str, default='test', help='train, val, test, etc')
        # self.parser.add_argument('--which_epoch', type=str, default='latest', help='which epoch to load? set to latest to use latest cached model')
        # self.parser.add_argument('--how_many', type=int, default=1, help='how many test images to run')
        self.parser.set_defaults(model='test')
        self.parser.set_defaults(how_many=1)
        self.parser.set_defaults(which_epoch='latest')
        self.isTrain = False



import PIL.Image as Image
import os
from torchvision import transforms as transforms

img_path_list = [
    "./visualization/origin/TUM-AEPINLNQ.png",
    "./visualization/origin/TUM-DFGFFNEY.png",
    "./visualization/origin/TUM-EWFNFSQL.png",
    "./visualization/origin/TUM-TCGA-CVATFAAT.png",
]

save_dir_path = "./visualization/stain_augmentation"
if not os.path.exists(save_dir_path):
    os.mkdir(save_dir_path)

if __name__ == "__main__":
    for img_path in img_path_list:
        image = transforms.ColorJitter(
            brightness=0.35, contrast=0.5, saturation=0.5, hue=0.5
        )(Image.open(img_path))
        save_img_path = save_dir_path + "/{}".format(img_path.split("/")[-1])
        image.save(save_img_path)


import os
import cv2
import numpy as np
import time
import argparse
import yaml
import json
import random
import copy
from skimage import color
from fitter import Fitter



class DatasetConfig:
    def __init__(self, data_dir, save_dir, dataset_name="", methods="", color_space="LAB", random=False, n=0):
        self.data_dir = data_dir
        self.save_dir = save_dir
        self.dataset_name = dataset_name
        self.methods = methods
        self.color_space = color_space
        self.random = random
        self.n = n

    def __str__(self):
        return (f"DatasetConfig(data_dir={self.data_dir}, save_dir={self.save_dir}, "
                f"dataset_name={self.dataset_name}, methods={self.methods}, "
                f"color_space={self.color_space}, random={self.random}, n={self.n})")





def getavgstd(image):
    avg = []
    std = []
    image_avg_l = np.mean(image[:, :, 0])
    image_std_l = np.std(image[:, :, 0])
    image_avg_a = np.mean(image[:, :, 1])
    image_std_a = np.std(image[:, :, 1])
    image_avg_b = np.mean(image[:, :, 2])
    image_std_b = np.std(image[:, :, 2])
    avg.append(image_avg_l)
    avg.append(image_avg_a)
    avg.append(image_avg_b)
    std.append(image_std_l)
    std.append(image_std_a)
    std.append(image_std_b)
    return (avg, std)


class RandStainPreprocess:

    
    def do(self, data_dir='', save_dir='', dataset_name="", methods="", color_space="LAB", random=False, n=0):
        args = DatasetConfig(data_dir, save_dir, dataset_name, methods, color_space, random, n)

        path_dataset = args.data_dir

        labL_avg_List = []
        labA_avg_List = []
        labB_avg_List = []
        labL_std_List = []
        labA_std_List = []
        labB_std_List = []

        t1 = time.time()
        i = 0

        
        if True:
            path_class_list = os.listdir(path_dataset)
            if args.random:
                random.shuffle(path_class_list)

            for image in path_class_list:
                if args.n == 0:  # n=0: all images each class
                    pass
                elif i < args.n:
                    i += 1
                else:
                    i = 0
                    break
                path_img = os.path.join(path_dataset, image)
                img = cv2.imread(path_img)
                try:  # debug
                    if args.color_space == "LAB":
                        img = cv2.cvtColor(img, cv2.COLOR_BGR2LAB)
                    elif args.color_space == "HED":
                        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                        img = color.rgb2hed(img)
                    elif args.color_space == "HSV":
                        img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
                    else:
                        print("wrong color space: {}!!".format(args.color_space))
                    img_avg, img_std = getavgstd(img)
                except:
                    continue
                    #print(path_img)
                labL_avg_List.append(img_avg[0])
                labA_avg_List.append(img_avg[1])
                labB_avg_List.append(img_avg[2])
                labL_std_List.append(img_std[0])
                labA_std_List.append(img_std[1])
                labB_std_List.append(img_std[2])
        t2 = time.time()
        print(t2 - t1)
        l_avg_mean = np.mean(labL_avg_List).item()
        l_avg_std = np.std(labL_avg_List).item()
        l_std_mean = np.mean(labL_std_List).item()
        l_std_std = np.std(labL_std_List).item()
        a_avg_mean = np.mean(labA_avg_List).item()
        a_avg_std = np.std(labA_avg_List).item()
        a_std_mean = np.mean(labA_std_List).item()
        a_std_std = np.std(labA_std_List).item()
        b_avg_mean = np.mean(labB_avg_List).item()
        b_avg_std = np.std(labB_avg_List).item()
        b_std_mean = np.mean(labB_std_List).item()
        b_std_std = np.std(labB_std_List).item()

        std_avg_list = [
            labL_avg_List,
            labL_std_List,
            labA_avg_List,
            labA_std_List,
            labB_avg_List,
            labB_std_List,
        ]
        distribution = []
        for std_avg in std_avg_list:
            f = Fitter(std_avg, distributions=["norm", "laplace"])
            f.fit()
            distribution.append(list(f.get_best(method="sumsquare_error").keys())[0])

        yaml_dict_lab = {
            "random": args.random,
            "n_each_class": args.n,
            "color_space": args.color_space,
            "methods": args.methods,
            "{}".format(args.color_space[0]): {  # lab-L/hed-H
                "avg": {
                    "mean": round(l_avg_mean, 3),
                    "std": round(l_avg_std, 3),
                    "distribution": distribution[0],
                },
                "std": {
                    "mean": round(l_std_mean, 3),
                    "std": round(l_std_std, 3),
                    "distribution": distribution[1],
                },
            },
            "{}".format(args.color_space[1]): {  # lab-A/hed-E
                "avg": {
                    "mean": round(a_avg_mean, 3),
                    "std": round(a_avg_std, 3),
                    "distribution": distribution[2],
                },
                "std": {
                    "mean": round(a_std_mean, 3),
                    "std": round(a_std_std, 3),
                    "distribution": distribution[3],
                },
            },
            "{}".format(args.color_space[2]): {  # lab-B/hed-D
                "avg": {
                    "mean": round(b_avg_mean, 3),
                    "std": round(b_avg_std, 3),
                    "distribution": distribution[4],
                },
                "std": {
                    "mean": round(b_std_mean, 3),
                    "std": round(b_std_std, 3),
                    "distribution": distribution[5],
                },
            },
        }
        yaml_save_path = "{}/{}.yaml".format(
            args.save_dir,
            args.dataset_name
            if args.dataset_name != ""
            else "dataset_{}_random{}_n{}".format(args.color_space, args.random, args.n),
        )
        with open(yaml_save_path, "w") as f:
            yaml.dump(yaml_dict_lab, f)
            print("The dataset lab statistics has been saved in {}".format(yaml_save_path))
        
        return yaml_save_path



def CreateDataLoader(opt , path_image):
    from ..data.custom_dataset_data_loader import CustomDatasetDataLoader
    data_loader = CustomDatasetDataLoader()
    #print(data_loader.name())
    data_loader.initialize(opt, path_image)
    return data_loader






import cv2
import numpy as np
from skimage import color
from typing import Optional, Dict
import yaml

# ToDo: add documents


class Dict2Class(object):
    # ToDo: Wrap into RandStainNA
    def __init__(self, my_dict: Dict):
        self.my_dict = my_dict
        for key in my_dict:
            setattr(self, key, my_dict[key])


def get_yaml_data(yaml_file):
    # ToDo: Wrap into RandStainNA
    file = open(yaml_file, "r", encoding="utf-8")
    file_data = file.read()
    file.close()
    # str->dict
    data = yaml.load(file_data, Loader=yaml.FullLoader)

    return data


class RandStainNAConvertor(object):
    # ToDo: support downloading yaml file from online if the path is not provided.
    def __init__(
        self,
        yaml_file: str,
        std_hyper: Optional[float] = 0,
        distribution: Optional[str] = "normal",
        probability: Optional[float] = 1.0,
        is_train: Optional[bool] = True,
    ):

        # true:training setting/false: demo setting

        assert distribution in [
            "normal",
            "laplace",
            "uniform",
        ], "Unsupported distribution style {}.".format(distribution)

        self.yaml_file = yaml_file
        cfg = get_yaml_data(self.yaml_file)
        c_s = cfg["color_space"]

        self._channel_avgs = {
            "avg": [
                cfg[c_s[0]]["avg"]["mean"],
                cfg[c_s[1]]["avg"]["mean"],
                cfg[c_s[2]]["avg"]["mean"],
            ],
            "std": [
                cfg[c_s[0]]["avg"]["std"],
                cfg[c_s[1]]["avg"]["std"],
                cfg[c_s[2]]["avg"]["std"],
            ],
        }
        self._channel_stds = {
            "avg": [
                cfg[c_s[0]]["std"]["mean"],
                cfg[c_s[1]]["std"]["mean"],
                cfg[c_s[2]]["std"]["mean"],
            ],
            "std": [
                cfg[c_s[0]]["std"]["std"],
                cfg[c_s[1]]["std"]["std"],
                cfg[c_s[2]]["std"]["std"],
            ],
        }

        self.channel_avgs = Dict2Class(self._channel_avgs)
        self.channel_stds = Dict2Class(self._channel_stds)

        self.color_space = cfg["color_space"]
        self.p = probability
        self.std_adjust = std_hyper
        self.color_space = c_s
        self.distribution = distribution
        self.is_train = is_train

    def _getavgstd(self, image: np.ndarray, isReturnNumpy: Optional[bool] = True):

        avgs = []
        stds = []

        num_of_channel = image.shape[2]
        for idx in range(num_of_channel):
            avgs.append(np.mean(image[:, :, idx]))
            stds.append(np.std(image[:, :, idx]))

        if isReturnNumpy:
            return (np.array(avgs), np.array(stds))
        else:
            return (avgs, stds)

    def _normalize(
        self,
        img: np.ndarray,
        img_avgs: np.ndarray,
        img_stds: np.ndarray,
        tar_avgs: np.ndarray,
        tar_stds: np.ndarray,
    ) -> np.ndarray:

        img_stds = np.clip(img_stds, 0.0001, 255)
        img = (img - img_avgs) * (tar_stds / img_stds) + tar_avgs

        if self.color_space in ["LAB", "HSV"]:
            img = np.clip(img, 0, 255).astype(np.uint8)

        return img

    def augment(self, img):
        # img:is_train:false——>np.array()(cv2.imread()) #BGR
        # img:is_train:True——>PIL.Image #RGB

        if self.is_train == False:
            image = img
        else:
            image = cv2.cvtColor(np.array(img), cv2.COLOR_RGB2BGR)

        num_of_channel = image.shape[2]

        # color space transfer
        if self.color_space == "LAB":
            image = cv2.cvtColor(image, cv2.COLOR_BGR2LAB)
        elif self.color_space == "HSV":
            image = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        elif self.color_space == "HED":
            image = color.rgb2hed(cv2.cvtColor(image, cv2.COLOR_BGR2RGB))

        std_adjust = self.std_adjust

        # virtual template generation
        tar_avgs = []
        tar_stds = []
        if self.distribution == "uniform":

            # three-sigma rule for uniform distribution
            for idx in range(num_of_channel):

                tar_avg = np.random.uniform(
                    low=self.channel_avgs.avg[idx] - 3 * self.channel_avgs.std[idx],
                    high=self.channel_avgs.avg[idx] + 3 * self.channel_avgs.std[idx],
                )
                tar_std = np.random.uniform(
                    low=self.channel_stds.avg[idx] - 3 * self.channel_stds.std[idx],
                    high=self.channel_stds.avg[idx] + 3 * self.channel_stds.std[idx],
                )

                tar_avgs.append(tar_avg)
                tar_stds.append(tar_std)
        else:
            if self.distribution == "normal":
                np_distribution = np.random.normal
            elif self.distribution == "laplace":
                np_distribution = np.random.laplace

            for idx in range(num_of_channel):
                tar_avg = np_distribution(
                    loc=self.channel_avgs.avg[idx],
                    scale=self.channel_avgs.std[idx] * (1 + std_adjust),
                )

                tar_std = np_distribution(
                    loc=self.channel_stds.avg[idx],
                    scale=self.channel_stds.std[idx] * (1 + std_adjust),
                )
                tar_avgs.append(tar_avg)
                tar_stds.append(tar_std)

        tar_avgs = np.array(tar_avgs)
        tar_stds = np.array(tar_stds)

        img_avgs, img_stds = self._getavgstd(image)

        image = self._normalize(
            img=image,
            img_avgs=img_avgs,
            img_stds=img_stds,
            tar_avgs=tar_avgs,
            tar_stds=tar_stds,
        )

        if self.color_space == "LAB":
            image = cv2.cvtColor(image, cv2.COLOR_LAB2BGR)
        elif self.color_space == "HSV":
            image = cv2.cvtColor(image, cv2.COLOR_HSV2BGR)
        elif self.color_space == "HED":
            nimg = color.hed2rgb(image)
            imin = nimg.min()
            imax = nimg.max()
            rsimg = (255 * (nimg - imin) / (imax - imin)).astype(
                "uint8"
            )  # rescale to [0,255]

            image = cv2.cvtColor(rsimg, cv2.COLOR_RGB2BGR)

        return image

    def __call__(self, img):
        if np.random.rand(1) < self.p:
            return self.augment(img)
        else:
            return img

    def __repr__(self):
        format_string = self.__class__.__name__ + "("
        format_string += f"methods=Reinhard"
        format_string += f", colorspace={self.color_space}"
        format_string += f", mean={self._channel_avgs}"
        format_string += f", std={self._channel_stds}"
        format_string += f", std_adjust={self.std_adjust}"
        format_string += f", distribution={self.distribution}"
        format_string += f", p={self.p})"
        return format_string


import numpy as np
import torch
from .StainNet.models import StainNet as STNet
import requests
from tqdm import tqdm
import os
from ..image_handler import LoadImages


class StainNet:
    
    # def __init__(self, mode):
        
    #     self.model_stainnet = StainNet()
    #     if mode == "cytopathology":
    #         self.model_stainnet.load_state_dict(torch.load("./StainNet/checkpoints/aligned_cytopathology_dataset/StainNet-3x0_best_psnr_layer3_ch32.pth", map_location=torch.device('cpu')))
    #     if mode == "histopathology":
    #         self.model_stainnet.load_state_dict(torch.load("./StainNet/checkpoints/aligned_histopathology_dataset/StainNet-Public_layer3_ch32.pth", map_location=torch.device('cpu')))
    #     if mode == "camelyon":
    #         self.model_stainnet.load_state_dict(torch.load("./StainNet/checkpoints/camelyon16_dataset/StainNet-Public-centerUni_layer3_ch32.pth", map_location=torch.device('cpu')))

    def __init__(self, model_path='',model_pathr='', run_type='cpu'):
        self.model_name = "stainet_model.pth"
        self.is_model_ready = model_path != ''
        self.model_path = model_path
        self.run_type = run_type
        self._check_model(self._get_safe_path(self.model_name))
        
        
    def _get_safe_path(self,filename):
        home_dir = os.path.expanduser("~")
        dir_name = ".lbbnorm"
        safe_path = os.path.join(home_dir, dir_name)
        if not os.path.exists(safe_path):
            os.makedirs(safe_path)
        return os.path.join(safe_path, filename)
    
    def _check_model(self,model_path):
        
        if os.path.exists(model_path):
            self.model_path = model_path
            self.is_model_ready = True
            return True
        else:
            return False
    

    def download_model(self):
        destination = self._get_safe_path(self.model_name)
        
        if self._check_model(destination):
            print('Model already exists.')
            return
        
        download_url = "http://s79.ir/lbb/stainnet_model.pth"
        response = requests.get(download_url, stream=True)
        total_size_in_bytes= int(response.headers.get('content-length', 0))
        block_size = 1024
        progress_bar = tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)
        with open(destination, 'wb') as file:
            for data in response.iter_content(block_size):
                progress_bar.update(len(data))
                file.write(data)
        progress_bar.close()
        if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
            print("ERROR, something went wrong, check your internet connection.")
            try:
                os.remove(destination)
            except:
                pass
        else:
            self.model_path = destination
            self.is_model_ready = True
  
    
    def norm(self, image):
        image = np.array(image).astype(np.float32)
        image = image.transpose((2, 0, 1))
        image = ((image / 255) - 0.5) / 0.5
        image=image[np.newaxis, ...]
        image=torch.from_numpy(image)
        return image

    def un_norm(self, image):
        image = image.cpu().detach().numpy()[0]
        image = ((image * 0.5 + 0.5) * 255).astype(np.uint8).transpose((1,2,0))
        return image   
    
    
    def transform(self, path_image):
        
        if not self._check_model(self.model_path) or len(self.model_path) < 1:
            print("""Pretrained model file not found, please download or specify it first.
Hint: you can download the pretrained model with `model.download_model()`, or you can specify it like this `StainGAN(model_path='../../../model.pth')` on creation time.""")
            FileNotFoundError()
            return
        print(self.model_path)
        self.model_stainnet = STNet()
        self.model_stainnet.load_state_dict(torch.load(self.model_path, map_location=torch.device(self.run_type)))

        images,_,__ = LoadImages(path_image)
        
        output = []
        for img in images:
            image_net=self.model_stainnet(self.norm(img))
            image_net=self.un_norm(image_net)
            
            output.append(image_net)
        
        
        if len(output) == 1:
            return output[0]
        else:
            return image_net



def create_model(opt,load_path):
    model = None
    #print(opt.model)
    if opt.model == 'cycle_gan':
        assert(opt.dataset_mode == 'unaligned')
        from .cycle_gan_model import CycleGANModel
        model = CycleGANModel()
    elif opt.model == 'test':
        assert(opt.dataset_mode == 'single')
        from .test_model import TestModel
        model = TestModel()
    else:
        raise ValueError("Model [%s] not recognized." % opt.model)
    model.initialize(opt,load_path)
    #print("model [%s] was created" % (model.name()))
    return model



from __future__ import print_function
import torch
import numpy as np
from PIL import Image
import inspect
import re
import numpy as np
import os
import collections


# Converts a Tensor into a Numpy array
# |imtype|: the desired type of the converted numpy array
def tensor2im(image_tensor, imtype=np.uint8):
    image_numpy = image_tensor[0].cpu().float().numpy()
    if image_numpy.shape[0] == 1:
        image_numpy = np.tile(image_numpy, (3, 1, 1))
    image_numpy = (np.transpose(image_numpy, (1, 2, 0)) + 1) / 2.0 * 255.0
    return image_numpy.astype(imtype)


def diagnose_network(net, name='network'):
    mean = 0.0
    count = 0
    for param in net.parameters():
        if param.grad is not None:
            mean += torch.mean(torch.abs(param.grad.data))
            count += 1
    if count > 0:
        mean = mean / count
    #print(name)
    #print(mean)


def save_image(image_numpy, image_path):
    image_pil = Image.fromarray(image_numpy)
    #image_pil.save(image_path)
    return image_pil


def print_numpy(x, val=True, shp=False):
    ...
    # x = x.astype(np.float64)
    # if shp:
    #     #print('shape,', x.shape)
    # if val:
    #     x = x.flatten()
    #     #print('mean = %3.3f, min = %3.3f, max = %3.3f, median = %3.3f, std=%3.3f' % (
    #     #    np.mean(x), np.min(x), np.max(x), np.median(x), np.std(x)))


def mkdirs(paths):
    if isinstance(paths, list) and not isinstance(paths, str):
        for path in paths:
            mkdir(path)
    else:
        mkdir(paths)


def mkdir(path):
    if not os.path.exists(path):
        os.makedirs(path)



from .base_options import BaseOptions


class TrainOptions(BaseOptions):
    def initialize(self):

        BaseOptions.initialize(self)
        self.parser.add_argument('--display_freq', type=int, default=100, help='frequency of showing training results on screen')
        self.parser.add_argument('--display_single_pane_ncols', type=int, default=0, help='if positive, display all images in a single visdom web panel with certain number of images per row.')
        self.parser.add_argument('--update_html_freq', type=int, default=1000, help='frequency of saving training results to html')
        self.parser.add_argument('--#print_freq', type=int, default=100, help='frequency of showing training results on console')
        self.parser.add_argument('--save_latest_freq', type=int, default=5000, help='frequency of saving the latest results')
        self.parser.add_argument('--save_epoch_freq', type=int, default=1, help='frequency of saving checkpoints at the end of epochs')
        self.parser.add_argument('--continue_train', action='store_true', help='continue training: load the latest model')
        self.parser.add_argument('--aicha_loss', action='store_true', help='use aicha loss or not')
        self.parser.add_argument('--epoch_count', type=int, default=1, help='the starting epoch count, we save the model by <epoch_count>, <epoch_count>+<save_latest_freq>, ...')
        self.parser.add_argument('--phase', type=str, default='train', help='train, val, test, etc')
        self.parser.add_argument('--which_epoch', type=str, default='latest', help='which epoch to load? set to latest to use latest cached model')
        self.parser.add_argument('--niter', type=int, default=100, help='# of iter at starting learning rate')
        self.parser.add_argument('--niter_decay', type=int, default=100, help='# of iter to linearly decay learning rate to zero')
        self.parser.add_argument('--beta1', type=float, default=0.5, help='momentum term of adam')
        self.parser.add_argument('--lr', type=float, default=0.0002, help='initial learning rate for adam')
        self.parser.add_argument('--no_lsgan', action='store_true', help='do *not* use least square GAN, if false, use vanilla GAN')
        self.parser.add_argument('--lambda_Aicha', type=float, default=10.0, help='weight for aicha loss (A -> B -> A)')
        self.parser.add_argument('--lambda_A', type=float, default=10.0, help='weight for cycle loss (A -> B -> A)')
        self.parser.add_argument('--lambda_B', type=float, default=10.0, help='weight for cycle loss (B -> A -> B)')
        self.parser.add_argument('--pool_size', type=int, default=50, help='the size of image buffer that stores previously generated images')
        self.parser.add_argument('--no_html', action='store_true', help='do not save intermediate training results to [opt.checkpoints_dir]/[opt.name]/web/')
        self.parser.add_argument('--lr_policy', type=str, default='lambda', help='learning rate policy: lambda|step|plateau')
        self.parser.add_argument('--lr_decay_iters', type=int, default=50, help='multiply by a gamma every lr_decay_iters iterations')
        self.parser.add_argument('--identity', type=float, default=0.5, help='use identity mapping. Setting identity other than 1 has an effect of scaling the weight of the identity mapping loss. For example, if the weight of the identity loss should be 10 times smaller than the weight of the reconstruction loss, please set optidentity = 0.1')

        self.isTrain = True



from setuptools import setup, find_packages


required_packages = ['numpy', 'tensorflow','opencv-python','spams-bin'
                     ,'scikit-image','scipy','torch',
                     'torchvision', 'dominate', 'visdom', 'pillow', 'visdom',
                     'imageio', 'tqdm', 'lmdb', 'staintools', 'fitter'
                     ]

setup(
    
    name='LBBNorm',
    version='1.3.0',
    packages=find_packages('src'),
    author='Laboratory of Systems Biology and Bioinformatics (LBB)',
    author_email='amasoudin@ut.ac.ir',
    description='Empower Your Tomorrow, Conquer the Future!',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    license='MIT',
    package_dir={'': 'src'},
    classifiers=[
        'Development Status :: 4 - Beta', 
        
        'Intended Audience :: Science/Research',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Information Technology',
        
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Topic :: Scientific/Engineering :: Image Recognition',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Operating System :: OS Independent',
    ],

    keywords='Healthcare Bio-Informatics',

    install_requires= required_packages,

)
